/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#include "nv-misc.h"
#include "os-interface.h"
#include "nv-linux.h"
#include "nv_compiler.h"
#include "os-agp.h"

#ifdef MODULE_ALIAS_CHARDEV_MAJOR
MODULE_ALIAS_CHARDEV_MAJOR(NV_MAJOR_DEVICE_NUMBER);
#endif

/*
 * our global state; one per device
 */

static int num_nv_devices = 0;

nv_linux_state_t nv_linux_devices[NV_MAX_DEVICES] = { { { 0 } } };

#if defined(NV_PM_SUPPORT_APM)
static struct pm_dev *apm_nv_dev[NV_MAX_DEVICES] = { 0 };
#endif

#if defined(NV_BUILD_NV_PAT_SUPPORT)
static int pat_enabled = 0;
#endif

/*
 * And one for the control device
 */

nv_linux_state_t nv_ctl_device = { { 0 } };
wait_queue_head_t nv_ctl_waitqueue;

// keep track of opened clients and their process id so they
//   can be free'd up on abnormal close
nv_client_t       nv_clients[NV_MAX_CLIENTS];

#ifdef CONFIG_PROC_FS
struct proc_dir_entry *proc_nvidia;
#endif

#ifdef CONFIG_DEVFS_FS
devfs_handle_t nv_devfs_handles[NV_MAX_DEVICES+1];
#endif

#ifdef NV_CLASS_SIMPLE_PRESENT
struct class_simple *class_nvidia;
#endif

#ifdef NV_SWIOTLB
int nv_swiotlb = 0;
#endif

// #define NV_DBG_MEM 1

// allow an easy way to convert all debug printfs related to memory
// management back and forth between 'info' and 'errors'
#if defined(NV_DBG_MEM)
#define NV_DBG_MEMINFO NV_DBG_ERRORS
#else
#define NV_DBG_MEMINFO NV_DBG_INFO
#endif

// allow an easy way to convert all debug printfs related to events
// back and forth between 'info' and 'errors'
#if defined(NV_DBG_EVENTS)
#define NV_DBG_EVENTINFO NV_DBG_ERRORS
#else
#define NV_DBG_EVENTINFO NV_DBG_INFO
#endif

// verify access to pci config space wasn't disabled behind our back
// unfortunately, XFree86 enables/disables memory access in pci config space at
// various times (such as restoring initial pci config space settings during vt
// switches or when doing mulicard). As a result, all of our register accesses
// are garbage at this point. add a check to see if access was disabled and
// reenable any such access.
#define NV_CHECK_PCI_CONFIG(nv) \
    nv_check_pci_config(nv, __LINE__)

static inline void nv_check_pci_config(nv_state_t *nv, int line)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);
    unsigned short cmd, flag = 0;

    // don't do this on the control device, only the actual devices
    if (nv->flags & NV_FLAG_CONTROL)
        return;

    pci_read_config_word(nvl->dev, PCI_COMMAND, &cmd);
    if (!(cmd & PCI_COMMAND_MASTER))
    {
        nv_printf(NV_DBG_USERERRORS, "NVRM: restoring bus mastering! (%d)\n", line);
        cmd |= PCI_COMMAND_MASTER;
        flag = 1;
    }

    if (!(cmd & PCI_COMMAND_MEMORY))
    {
        nv_printf(NV_DBG_USERERRORS, "NVRM: restoring MEM access! (%d)\n", line);
        cmd |= PCI_COMMAND_MEMORY;
        flag = 1;
    }

    if (flag)
        pci_write_config_word(nvl->dev, PCI_COMMAND, cmd);
}

/***
 *** STATIC functions, only in this file
 ***/

/* nvos_ functions.. do not take a state device parameter  */
static int      nvos_post_vbios(nv_ioctl_post_vbios_t *info);
static void     nvos_proc_create(void);
static void     nvos_proc_remove_all(struct proc_dir_entry *);
static void     nvos_proc_remove(void);
static int      nvos_malloc_pages(nv_alloc_t *);
static void     nvos_unlock_pages(nv_alloc_t *);
static void     nvos_free_pages(nv_alloc_t *);
static int      nvos_count_devices(void);

#define nvos_unlock_and_free_pages(at_count, at) \
    if (at->page_table) {                        \
        if (at_count == 0)                       \
            nvos_unlock_pages(at);               \
        nvos_free_pages(at);                     \
    }

static nv_alloc_t  *nvos_create_alloc(struct pci_dev *, int);
static int          nvos_free_alloc(nv_alloc_t *);

/* nvl_ functions.. take a linux state device pointer */
static nv_alloc_t  *nvl_find_alloc(nv_linux_state_t *, unsigned long, unsigned long);
static int          nvl_add_alloc(nv_linux_state_t *, nv_alloc_t *);
static int          nvl_remove_alloc(nv_linux_state_t *, nv_alloc_t *);

/* lock-related functions that should only be called from this file */
static void nv_lock_init_locks(nv_state_t *nv);

#define nv_init_lock(lock)  spin_lock_init(&lock)
#define nv_lock(lock)       spin_lock(&lock)
#define nv_unlock(lock)     spin_unlock(&lock)
#define nv_down(lock)       down(&lock)
#define nv_up(lock)         up(&lock)

#define nv_lock_irq(lock,flags)    spin_lock_irqsave(&lock,flags)
#define nv_unlock_irq(lock,flags)  spin_unlock_irqrestore(&lock,flags)


/***
 *** EXPORTS to Linux Kernel
 ***/

/* nv_kern_ functions, interfaces used by linux kernel */
void          nv_kern_vma_open(struct vm_area_struct *vma);
void          nv_kern_vma_release(struct vm_area_struct *vma);

int           nv_kern_open(struct inode *, struct file *);
int           nv_kern_close(struct inode *, struct file *);
int           nv_kern_mmap(struct file *, struct vm_area_struct *);
unsigned int  nv_kern_poll(struct file *, poll_table *);
int           nv_kern_ioctl(struct inode *, struct file *, unsigned int, unsigned long);
void          nv_kern_isr_bh(unsigned long);
irqreturn_t   nv_kern_isr(int, void *, struct pt_regs *);
void          nv_kern_rc_timer(unsigned long);
#if defined(NV_PM_SUPPORT_APM)
int           nv_kern_apm_event(struct pm_dev *dev, pm_request_t rqst, void *data);
#endif

int           nv_kern_read_cardinfo(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_status(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_agpinfo(char *, char **, off_t off, int, int *, void *);
int           nv_kern_read_version(char *, char **, off_t off, int, int *, void *);

int           nv_kern_ctl_open(struct inode *, struct file *);
int           nv_kern_ctl_close(struct inode *, struct file *);
unsigned int  nv_kern_ctl_poll(struct file *, poll_table *);

int nv_kern_probe(struct pci_dev *, const struct pci_device_id *);
#if defined(NV_PM_SUPPORT_ACPI)
int nv_kern_acpi_standby(struct pci_dev *, u32);
int nv_kern_acpi_resume(struct pci_dev *);
#endif

/***
 *** see nv.h for functions exported to other parts of resman
 ***/

static struct pci_device_id nv_pci_table[] = {
    { 
        .vendor      = PCI_VENDOR_ID_NVIDIA,
        .device      = PCI_ANY_ID,
        .subvendor   = PCI_ANY_ID,
        .subdevice   = PCI_ANY_ID,
        .class       = (PCI_CLASS_DISPLAY_VGA << 8),
        .class_mask  = ~0,
    },
    { }
};

MODULE_DEVICE_TABLE(pci, nv_pci_table);

static struct pci_driver nv_pci_driver = {
    .name     = "nvidia",
    .id_table = nv_pci_table,
    .probe    = nv_kern_probe,
#if defined(NV_PM_SUPPORT_ACPI)
    .suspend  = nv_kern_acpi_standby,
    .resume   = nv_kern_acpi_resume,
#endif
};

/* character driver entry points */

static struct file_operations nv_fops = {
    .owner     = THIS_MODULE,
    .poll      = nv_kern_poll,
    .ioctl     = nv_kern_ioctl,
    .mmap      = nv_kern_mmap,
    .open      = nv_kern_open,
    .release   = nv_kern_close,
};

// Our reserved major device number.
int nv_major = NV_MAJOR_DEVICE_NUMBER;

// pull in the pointer to the NVID stamp from the binary module
extern const char *pNVRM_ID;

#if NV_ENABLE_MEM_TRACKING
// poor man's memory tracker, see nv-linux.h
int vm_usage = 0;
int km_usage = 0;
int fp_usage = 0;

struct mem_track_t *vm_list = NULL;
struct mem_track_t *km_list = NULL;
struct mem_track_t *fp_list = NULL;
#endif /* NV_ENABLE_MEM_TRACKING */

/***
 *** STATIC functions
 ***/

/* specify that this card needs it's vbios posted */
static int nvos_post_vbios(nv_ioctl_post_vbios_t *info)
{
    int i;

    for (i = 0; i < NV_MAX_DEVICES; i++)
    {
        nv_state_t *nv = NV_STATE_PTR(&nv_linux_devices[i]);
        if (nv->bus == info->bus && nv->slot == info->slot)
        {
            // we assume any device was already posted and rely on
            // X to tell us which cards need posting. But if we've
            // already manually posted a card, it doesn't need to 
            // be reposted again.
            if (!(nv->flags & NV_FLAG_WAS_POSTED))
            {
                nv->flags |= NV_FLAG_NEEDS_POSTING;
            }
        }
    }

    return 0;
}


#if defined(NV_DBG_MEM)
static void
nvos_list_page_count(nv_pte_t *page_list, unsigned long num_pages)
{
    nv_pte_t *page_ptr = page_list;

    if (page_ptr == NULL)
        return;

    nv_printf(NV_DBG_MEMINFO, "NVRM:  page_table at 0x%p, %d pages\n", 
        page_ptr, num_pages);

    while (num_pages)
    {
        nv_printf(NV_DBG_MEMINFO, "NVRM:   0x%x (0x%x): count %d flags 0x%x\n", 
            page_ptr->phys_addr, page_ptr->dma_addr,
            (NV_GET_PAGE_STRUCT(page_ptr->phys_addr))->count,
            (NV_GET_PAGE_STRUCT(page_ptr->phys_addr))->flags);
        num_pages--;
        page_ptr++;
    }
}
#else
#define nvos_list_page_count(page_ptr, num_pages)
#endif

/*
 * Function to wrap CACHE_FLUSH so we can pass it to smp_call_function
 */

static void cache_flush(void *p)
{
    CACHE_FLUSH();
}

/*
 * The idea behind nvos_malloc is to manage physical memory for DMA buffers
 * directly (as opposed to using "vmalloc") to avoid exhausting the kernel's
 * virtual address space which can shrink down to 128MB on i386 systems with
 * large amounts of physical memory. This also minimizes fragmentation of
 * virtual memory.
 * While this isn't really necessary on ia64, it can't hurt either.
 */

/*
 * When called with __GFP_DMA, __get_free_pages should return physical pages
 * addressable by 32bit PCI hardware on the ia64. On ia32, this will always
 * be the case unless we explicitly tell __get_free_pages to give us "high"
 * memory on systems with high memory support enabled.
 */

/* note that there's a subtle kernel interaction with regards to bookkeeping
 * on these pages. So long as the pages are marked reserved, the kernel won't
 * touch them (alter the usage count on them). this leads to a subtle problem
 * with mmap. Normally, allocating the pages would set the count to 1, then 
 * mmaping them would bump the count up to 2. The order of unmapping and freeing
 * the pages wouldn't matter, as they wouldn't really be considered free by the
 * kernel until the count dropped back to 0. Since the kernel won't touch the
 * count when the page is reserved, we need to be careful about this order and
 * unreserving the pages. if we unreserve the pages while freeing them, and the
 * munmap comes later, the munmap code path will attempt a second free on the 
 * same pages. We also don't have a lot of control over which comes first, 
 * sometimes we'll get called to free the pages first, sometimes we'll get called
 * to munmap them first. Oh, and we'll get vma open/close calls every time the
 * process is cloned, then execv'd, and munmap == vma close.
 * sooo, we keep our own count of the allocation usage, and don't unreserve the
 * pages until our count drops to 0. this should currently happen in either
 * vma_release or nvos_free, both of which will be followed by a kernel attempt
 * to free the page. Since the page fill finally be unreserved, the kernel will
 * reduce the count to 0 and successfully free the page for us, only once.
 */

static int nvos_malloc_pages(
    nv_alloc_t       *at
)
{
    /* point page_ptr at the start of the actual page list */
    nv_pte_t *page_ptr = at->page_table;
    unsigned int i;

    nv_printf(NV_DBG_MEMINFO, "NVRM: nvos_malloc_pages: %d pages\n",
        at->num_pages);
    nv_printf(NV_DBG_MEMINFO, "NVRM:    page_table: 0x%x\n", page_ptr);

    for (i = 0; i < at->num_pages; i++) 
    {
        unsigned long virt_addr;
        unsigned long phys_addr;

        NV_GET_FREE_PAGES(virt_addr, 0); 
        if (virt_addr == 0) {
            nv_printf(NV_DBG_ERRORS, "NVRM: nvos_malloc failed to allocate a page\n");
            goto failed;
        }
        phys_addr = virt_to_phys((void *) virt_addr);

        /* lock the page for dma purposes */
        SetPageReserved(NV_GET_PAGE_STRUCT(phys_addr));

        page_ptr->phys_addr = phys_addr;
        page_ptr->virt_addr = virt_addr;
        page_ptr->dma_addr = page_ptr->phys_addr;

        NV_SET_PAGE_ATTRIB_UNCACHED(page_ptr);

#if defined(NV_SG_MAP_BUFFERS)
        // for amd 64-bit platforms, remap pages to make them 32-bit addressable
        // we'll only be called here for pci memory (not agp), so we always want
        // to do this. may eventually need to do this for other 64-bit platforms 
        // special note: as of 2.4.2x kernels, pci_map_sg calls pci_map_single,
        // which relies on __pa() to convert virtual -> physical addresses.
        // additionally, pci_map_single relies on the pages being physically 
        // contiguous. thankfully, this ties in nicely with our previous work-arounds
        // here, but causes other problems below in nv_alloc_pages.
        if (nv_sg_map_buffer(at->dev, &at->sg_list[i], 
                             (void *) page_ptr->virt_addr, 1))
        {
            nv_printf(NV_DBG_ERRORS, "NVRM: nvos_malloc failed to sg map pages\n");
            goto failed;
        }
        nv_sg_load(&at->sg_list[i], page_ptr, 1);
#endif
        if (i < 5)
            nv_printf(NV_DBG_MEMINFO, "   0x%lx (0x%lx) 0x%lx (0x%lx) 0x%lx\n", page_ptr->virt_addr, virt_addr, page_ptr->phys_addr, phys_addr, page_ptr->dma_addr);

        page_ptr++;
    }

    nvos_list_page_count(at->page_table, at->num_pages);

    // Flush the CPU caches.  Drivers may be using a write combined
    // mapping of these newly allocated pages.  If there are any
    // cached writes outstanding we must insure they reach memory
    // before any WC writes. We've seen instances of data corruption
    // due to this. the corruption issues appear to go away, starting
    // with kernel 2.4.23, but we've yet to isolate what change caused
    // the problems to go away, and whether the problem is really
    // fixed or the timing is just different. for that reason, we'll
    // enable this flush for all kernels, rather than just < 2.4.23
    // (at least until we can do more investigation).
#ifdef CONFIG_SMP
        smp_call_function(cache_flush, NULL, 1, 1);
#endif
        cache_flush(NULL);

    return 0;

failed:
    while (page_ptr != at->page_table)
    {
        // if we failed when allocating this page, skip over it
        // but if we failed pci_map_sg, make sure to free this page
        if (page_ptr->virt_addr)
        {
#if defined(NV_SG_MAP_BUFFERS)
            nv_sg_unmap_buffer(at->dev, &at->sg_list[i], page_ptr, 1);
#endif
            NV_UNLOCK_PAGE(page_ptr);
            NV_SET_PAGE_ATTRIB_CACHED(page_ptr);
            NV_FREE_PAGES(page_ptr->virt_addr, 0);
        }
        page_ptr--;
        i--;
    }

    return -1;
}

// unlock the pages we've locked down for dma purposes
static void nvos_unlock_pages(
    nv_alloc_t       *at
)
{
    nv_pte_t *page_ptr = at->page_table;
    int pages_left = at->num_pages;

    if (page_ptr== NULL)
        return;

    nv_printf(NV_DBG_MEMINFO, "NVRM: nvos_unlock_pages: %d pages\n",
        pages_left);

    while (pages_left)
    {
        NV_UNLOCK_PAGE(page_ptr);
        page_ptr++;
        pages_left--;
    }
}

static void nvos_free_pages(
    nv_alloc_t       *at
)
{
    nv_pte_t *page_ptr = at->page_table;
    unsigned long pages_left = at->num_pages;
    unsigned int i = 0;

    if (page_ptr == NULL)
        return;

    nv_printf(NV_DBG_MEMINFO, "NVRM: nvos_free_pages: %d pages\n", pages_left);

    nvos_list_page_count(at->page_table, at->num_pages);

    while (pages_left)
    {
        if (i < 5)
            nv_printf(NV_DBG_MEMINFO, "   0x%lx 0x%lx 0x%lx\n", page_ptr->virt_addr, page_ptr->phys_addr, page_ptr->dma_addr);
#if defined(NV_SG_MAP_BUFFERS)
        nv_sg_unmap_buffer(at->dev, &at->sg_list[i], page_ptr, 1);
#endif
        NV_SET_PAGE_ATTRIB_CACHED(page_ptr);
        NV_FREE_PAGES(page_ptr->virt_addr, 0);
        page_ptr++;
        i++;
        pages_left--;
    }
}

static 
nv_alloc_t *nvos_create_alloc(
    struct pci_dev *dev,
    int num_pages
)
{
    nv_alloc_t *at;
    int pt_size;

    NV_KMALLOC(at, sizeof(nv_alloc_t));
    if (at == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate alloc_t\n");
        return NULL;
    }

    memset(at, 0, sizeof(nv_alloc_t));

    pt_size = num_pages *  sizeof(nv_pte_t);
    NV_VMALLOC(at->page_table, pt_size);
    if (at->page_table == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate page table\n");
        NV_KFREE(at, sizeof(nv_alloc_t));
        return NULL;
    }
    memset(at->page_table, 0, pt_size);
    at->num_pages = num_pages;
    NV_ATOMIC_SET(at->usage_count, 0);

#if defined(NV_SG_MAP_BUFFERS)
    at->dev = dev;
    pt_size = num_pages * sizeof(struct scatterlist);
    NV_VMALLOC(at->sg_list, pt_size);
    if (at->sg_list == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate scatter gather list\n");
        NV_VFREE(at->page_table, pt_size);
        NV_KFREE(at, sizeof(nv_alloc_t));
        return NULL;
    }
    memset(at->sg_list, 0, pt_size);
#endif

    return at;
}

static 
int nvos_free_alloc(
    nv_alloc_t *at
)
{
    if (at == NULL)
        return -1;

    if (NV_ATOMIC_READ(at->usage_count))
        return 1;

    // we keep the page_table around after freeing the pages
    // for bookkeeping reasons. Free the page_table and assume
    // the underlying pages are already unlocked and freed.
    if (at->page_table)
        NV_VFREE(at->page_table, at->num_pages * sizeof(nv_pte_t));

#if defined(NV_SG_MAP_BUFFERS)
    if (at->sg_list)
        NV_VFREE(at->sg_list, at->num_pages * sizeof(struct scatterlist));
#endif

    NV_KFREE(at, sizeof(nv_alloc_t));

    return 0;
}

static u8 nvos_find_agp_capability(struct pci_dev *dev)
{
    u16 status;
    u8  cap_ptr, cap_id;

    pci_read_config_word(dev, PCI_STATUS, &status);
    status &= PCI_STATUS_CAP_LIST;
    if (!status)
        return 0;

    switch (dev->hdr_type) {
        case PCI_HEADER_TYPE_NORMAL:
        case PCI_HEADER_TYPE_BRIDGE:
            pci_read_config_byte(dev, PCI_CAPABILITY_LIST, &cap_ptr);
            break;
        default:
            return 0;
    }

    do {
        cap_ptr &= 0xfc;
        pci_read_config_byte(dev, cap_ptr + PCI_CAP_LIST_ID, &cap_id);
        if (cap_id == PCI_CAP_ID_AGP)
            return cap_ptr;
        pci_read_config_byte(dev, cap_ptr + PCI_CAP_LIST_NEXT, &cap_ptr);
    } while (cap_ptr && cap_id != 0xff);

    return 0;
}

static u8 nvos_find_pci_express_capability(struct pci_dev *dev)
{
    u16 status;
    u8  cap_ptr, cap_id;

    pci_read_config_word(dev, PCI_STATUS, &status);
    status &= PCI_STATUS_CAP_LIST;
    if (!status)
        return 0;

    switch (dev->hdr_type) {
        case PCI_HEADER_TYPE_NORMAL:
        case PCI_HEADER_TYPE_BRIDGE:
            pci_read_config_byte(dev, PCI_CAPABILITY_LIST, &cap_ptr);
            break;
        default:
            return 0;
    }

    do {
        cap_ptr &= 0xfc;
        pci_read_config_byte(dev, cap_ptr + PCI_CAP_LIST_ID, &cap_id);
        if (cap_id == PCI_CAP_ID_EXP)
            return cap_ptr;
        pci_read_config_byte(dev, cap_ptr + PCI_CAP_LIST_NEXT, &cap_ptr);
    } while (cap_ptr && cap_id != 0xff);

    return 0;
}

static struct pci_dev* nvos_find_agp_by_class(unsigned int class)
{
    struct pci_dev *dev, *fn;
    u8     cap_ptr;
    int    func, slot;

    dev = pci_find_class(class << 8, NULL);
    do {
        for (func = 0; func < 8; func++) {
            slot = PCI_SLOT(dev->devfn);
            fn = pci_find_slot(dev->bus->number, PCI_DEVFN(slot, func));
            if (!fn)
                continue;
            cap_ptr = nvos_find_agp_capability(fn);
            if (cap_ptr)
                return fn;
        }
        dev = pci_find_class(class << 8, dev);
    } while (dev);

    return NULL;
}

static struct pci_dev* nv_find_pci_dev(nv_state_t *nv)
{
    struct pci_dev *dev;
    int    bus  = 0;
    int    slot = 0;
    
    dev = NULL;
    do {
        dev = pci_find_device(nv->vendor_id, nv->device_id, dev);
        if (dev) {
            bus = dev->bus->number;
            slot = PCI_SLOT(dev->devfn);
        }
    } while (dev && (bus != nv->bus || slot != nv->slot));

    return dev;
}

static void nvos_proc_create(void)
{
#ifdef CONFIG_PROC_FS
    struct pci_dev *dev;
    int i = 0;
    char name[6];

    struct proc_dir_entry *entry;
    struct proc_dir_entry *proc_nvidia_agp, *proc_nvidia_cards;

    /* world readable directory */
    int flags = S_IFDIR | S_IRUGO | S_IXUGO;

    nv_state_t *nv;
    nv_linux_state_t *nvl;
    nv_linux_state_t *nv_max_devices;

    proc_nvidia = create_proc_entry("nvidia", flags, proc_root_driver);
    if (!proc_nvidia)
        goto failed;

    proc_nvidia_cards = create_proc_entry("cards", flags, proc_nvidia);
    if (!proc_nvidia_cards)
        goto failed;

    proc_nvidia_agp = create_proc_entry("agp", flags, proc_nvidia);
    if (!proc_nvidia_agp)
        goto failed;

    /*
     * Set the module owner to ensure that the reference
     * count reflects accesses to the proc files.
     */
    proc_nvidia->owner       = THIS_MODULE;
    proc_nvidia_cards->owner = THIS_MODULE;
    proc_nvidia_agp->owner   = THIS_MODULE;

    nv_max_devices = nv_linux_devices + NV_MAX_DEVICES;
    for (nvl = nv_linux_devices; nvl < nv_max_devices; nvl++) 
    {
        nv = NV_STATE_PTR(nvl);

        if (nv->device_id == 0)
            break;

        /* world readable file */
        flags = S_IFREG | S_IRUGO;

        dev = nv_find_pci_dev(nv);
        if (!dev)
            break;

        sprintf(name, "%d", i++);
        entry = create_proc_entry(name, flags, proc_nvidia_cards);
        if (!entry)
            goto failed;

        entry->data = nv;
        entry->read_proc = nv_kern_read_cardinfo;
        entry->owner = THIS_MODULE;

        if (nvos_find_agp_capability(dev)) {
            entry = create_proc_entry("status", flags, proc_nvidia_agp);
            if (!entry)
                goto failed;

            entry->data = nv;
            entry->read_proc = nv_kern_read_status;
            entry->owner = THIS_MODULE;

            entry = create_proc_entry("card", flags, proc_nvidia_agp);
            if (!entry)
                goto failed;

            entry->data = nv;
            entry->read_proc = nv_kern_read_agpinfo;
            entry->owner = THIS_MODULE;
        }
    }

    entry = create_proc_entry("version", flags, proc_nvidia);
    if (!entry)
        goto failed;

    entry->read_proc = nv_kern_read_version;
    entry->owner = THIS_MODULE;

    entry = create_proc_entry("host-bridge", flags, proc_nvidia_agp);
    if (!entry)
        goto failed;

    entry->data = NULL;
    entry->read_proc = nv_kern_read_agpinfo;
    entry->owner = THIS_MODULE;

    return;

failed:
    nv_printf(NV_DBG_ERRORS, "NVRM: failed to create /proc entries!\n");
    nvos_proc_remove_all(proc_nvidia);
#endif
}

#ifdef CONFIG_PROC_FS
static void nvos_proc_remove_all(struct proc_dir_entry *entry)
{
    while (entry) {
        struct proc_dir_entry *next = entry->next;
        if (entry->subdir)
            nvos_proc_remove_all(entry->subdir);
        remove_proc_entry(entry->name, entry->parent);
        entry = next;
    }
}
#endif

static void nvos_proc_remove(void)
{
#ifdef CONFIG_PROC_FS
    nvos_proc_remove_all(proc_nvidia);
#endif
}

/*
 * Given a virtual address, fid the 'at' that owns it
 * Uses the physical address as the key.
 */
static nv_alloc_t *nvl_find_alloc(
    nv_linux_state_t    *nvl,
    unsigned long  address,
    unsigned long  flags
)
{
    nv_alloc_t *at;

    for (at = nvl->alloc_queue; at; at = at->next)
    {
        // make sure this 'at' matches the flags the caller provided
        // ie, don't mistake a pci allocation with an agp allocation
        if (!(at->flags & flags))
            continue;

        // most mappings will be found based on the 'key'
        if (address == ((unsigned long) at->key_mapping))
            return at;

        if (at->page_table)
        {
            int i;
            for (i = 0; i < at->num_pages; i++)
            {
                unsigned long offset = at->page_table[i].phys_addr;
                if ((address >= offset) &&
                    (address < (offset + PAGE_SIZE)))
                    return at;
            }
        }

    }

    /* failure is not necessarily an error if the caller
       was just probing an address */
    nv_printf(NV_DBG_INFO, "NVRM: could not find map for vm 0x%lx\n", address);
    return NULL;
}

static int nvl_add_alloc(
    nv_linux_state_t *nvl, 
    nv_alloc_t *at
)
{
    nv_down(nvl->at_lock);
    at->next = nvl->alloc_queue;
    nvl->alloc_queue = at;
    nv_up(nvl->at_lock);
    return 0;
}

static int nvl_remove_alloc(
    nv_linux_state_t *nvl, 
    nv_alloc_t *at
)
{
    nv_alloc_t *tmp, *prev;

    if (nvl->alloc_queue == at)
    {
        nvl->alloc_queue = nvl->alloc_queue->next;
        return 0;
    }

    for (tmp = prev = nvl->alloc_queue; tmp; prev = tmp, tmp = tmp->next)
    {
        if (tmp == at)
        {
            prev->next = tmp->next;
            return 0;
        }
    }

    return -1;
}

#if defined(NV_BUILD_NV_PAT_SUPPORT)
/*
 * Private PAT support for use by the NVIDIA driver. This is an
 * interim solution until the kernel offers PAT support.
 */
static int   __check_pat_support       (void);
static void  __nv_setup_pat_entries    (void *);
static void  __nv_restore_pat_entries  (void *);
static int   __nv_enable_pat_support   (void);
static void  __nv_disable_pat_support  (void);

#define NV_READ_PAT_ENTRIES(pat1, pat2)   rdmsr(IA32_CR_PAT, (pat1), (pat2))
#define NV_WRITE_PAT_ENTRIES(pat1, pat2)  wrmsr(IA32_CR_PAT, (pat1), (pat2))
#define NV_PAT_ENTRY(pat, index)          (((pat) & (0xff<<((index)*8)))>>((index)*8))

static inline void __nv_disable_caches(unsigned long *cr4)
{
    unsigned long cr0 = read_cr0();
    write_cr0(((cr0 & (0xdfffffff)) | 0x40000000));
    wbinvd();
    *cr4 = read_cr4();
    if (*cr4 & 0x80) write_cr4(*cr4 & ~0x80);
    __flush_tlb();
}

static inline void __nv_enable_caches(unsigned long cr4)
{
    unsigned long cr0 = read_cr0();
    wbinvd();
    __flush_tlb();
    write_cr0((cr0 & 0x9fffffff));
    if (cr4 & 0x80) write_cr4(cr4);
}

static int __check_pat_support()
{
    unsigned int pat1, pat2, i;

    if (!test_bit(X86_FEATURE_PAT, (volatile unsigned long *)&boot_cpu_data.x86_capability))
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: cpu does not support PAT, aborting..\n");
        return 0;
    }

    NV_READ_PAT_ENTRIES(pat1, pat2);

    for (i = 0; i < 4; i++)
    {
         if (NV_PAT_ENTRY(pat1, i) == 1)
         {
             nv_printf(NV_DBG_ERRORS, "NVRM: PAT index %d already configured for Write-Combining!\n", i);
             nv_printf(NV_DBG_ERRORS, "NVRM: Aborting, due to PAT already being configured\n");
             return 0;
         }
    }

    for (i = 0; i < 4; i++)
    {
         if (NV_PAT_ENTRY(pat2, i) == 1)
         {
             nv_printf(NV_DBG_ERRORS, "NVRM: PAT index %d already configured for Write-Combining!\n", i + 4);
             nv_printf(NV_DBG_ERRORS, "NVRM: Aborting, due to PAT already being configured\n");
             return 0;
         }
    }

    return 1;
}

static unsigned long orig_pat1, orig_pat2;

static void __nv_setup_pat_entries(void *info)
{
    unsigned long pat1, pat2, cr4;
    unsigned long eflags;

    NV_SAVE_FLAGS(eflags);
    NV_CLI();
    __nv_disable_caches(&cr4);

    NV_READ_PAT_ENTRIES(pat1, pat2);

    pat1 &= 0xffff00ff;
    pat1 |= 0x00000100;

    NV_WRITE_PAT_ENTRIES(pat1, pat2);

    __nv_enable_caches(cr4);
    NV_RESTORE_FLAGS(eflags);
}

static void __nv_restore_pat_entries(void *info)
{
    unsigned long cr4;
    unsigned long eflags;

    NV_SAVE_FLAGS(eflags);
    NV_CLI();
    __nv_disable_caches(&cr4);

    NV_WRITE_PAT_ENTRIES(orig_pat1, orig_pat2);

    __nv_enable_caches(cr4);
    NV_RESTORE_FLAGS(eflags);
}

static int __nv_enable_pat_support()
{
    unsigned long pat1, pat2;

    if (pat_enabled)
        return 1;

    if (!__check_pat_support())
        return 0;

    NV_READ_PAT_ENTRIES(orig_pat1, orig_pat2);
    nv_printf(NV_DBG_SETUP, "saved orig pats as 0x%lx 0x%lx\n", orig_pat1, orig_pat2);

#ifdef CONFIG_SMP
    if (smp_call_function(__nv_setup_pat_entries, NULL, 1, 1) != 0)
        return 0;
#endif

    __nv_setup_pat_entries(NULL);

    pat_enabled = 1;

    NV_READ_PAT_ENTRIES(pat1, pat2);
    nv_printf(NV_DBG_SETUP, "changed pats to 0x%lx 0x%lx\n", pat1, pat2);

    return 1;
}

static void __nv_disable_pat_support()
{
    unsigned long pat1, pat2;

    if (!pat_enabled)
        return;

#ifdef CONFIG_SMP
    if (smp_call_function(__nv_restore_pat_entries, NULL, 1, 1) != 0)
        return;
#endif

    __nv_restore_pat_entries(NULL);

    pat_enabled = 0;

    NV_READ_PAT_ENTRIES(pat1, pat2);
    nv_printf(NV_DBG_SETUP, "restored orig pats as 0x%lx 0x%lx\n", pat1, pat2);
}

#endif /* defined(NV_BUILD_NV_PAT_SUPPORT) */

/***
 *** EXPORTS to Linux Kernel
 ***/

static int __init nvidia_init_module(void)
{
    int rc, i, count;

    memset(nv_linux_devices, 0, sizeof(nv_linux_devices));

    if (pci_register_driver(&nv_pci_driver) < 0)
    {
        pci_unregister_driver(&nv_pci_driver); // XXX ???
        nv_printf(NV_DBG_ERRORS, "NVRM: no NVIDIA graphics adapter found\n");
        return -ENODEV;
    }

    count = nvos_count_devices();
    if (num_nv_devices != count)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: the NVIDIA probe routine was not "
            "called for %d device(s)!!\n", count - num_nv_devices);
        if (!num_nv_devices)
        {
            nv_printf(NV_DBG_ERRORS, "NVRM: no devices probed, aborting!\n");
            nv_printf(NV_DBG_ERRORS, "NVRM: this often occurs when rivafb is "
                "loaded and claims the device's resources.\n");
            nv_printf(NV_DBG_ERRORS, "NVRM: try removing the rivafb module "
                "(or reconfiguring your kernel to remove\n");
            nv_printf(NV_DBG_ERRORS, "NVRM: rivafb support) and then try "
                "loading the NVIDIA kernel module again.\n");
            pci_unregister_driver(&nv_pci_driver);
            return -ENODEV;
        }
    }

    nv_printf(NV_DBG_ERRORS, "NVRM: loading %s\n", pNVRM_ID);

    rc = NV_REGISTER_CHRDEV(nv_major, "nvidia", &nv_fops);
    if (rc < 0)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: register chrdev failed\n");
        pci_unregister_driver(&nv_pci_driver);
        return rc;
    }

#ifdef NV_CLASS_SIMPLE_PRESENT
    class_nvidia = class_simple_create(THIS_MODULE, "nvidia");
    if (IS_ERR(class_nvidia))
    {
        rc = PTR_ERR(class_nvidia);
        nv_printf(NV_DBG_ERRORS, "NVRM: class_simple creation failed\n");
        /* goto failed; */
    }
#endif

#ifdef CONFIG_DEVFS_FS
    do
    {
        char name[10];

        nv_devfs_handles[0] = NV_DEVFS_REGISTER("nvidiactl", 255);

        for (i = 0; i < num_nv_devices; i++)
        {
            sprintf(name, "nvidia%d", i);
            nv_devfs_handles[i+1] = NV_DEVFS_REGISTER(name, i);
        }

        for (i = 0; i <= num_nv_devices; i++)
        {
            if (nv_devfs_handles[i] == NULL)
            {
                rc = -ENOMEM; /* XXX Fix me? (rc) */
                nv_printf(NV_DBG_ERRORS, "NVRM: devfs register failed\n");
                goto failed;
            }
        }
    } while(0);
#endif

#ifdef NV_CLASS_SIMPLE_PRESENT
    if (!IS_ERR(class_nvidia))
    {
        /*
         * XXX We don't consider class_simple errors fatal for
         * now to avoid unexpected failures.
         */
        char name[10];

        class_simple_device_add(class_nvidia,
            MKDEV(NV_MAJOR_DEVICE_NUMBER, 255), NULL, "nvidiactl");

        for (i = 0; i < num_nv_devices; i++)
        {
            sprintf(name, "nvidia%d", i);
            class_simple_device_add(class_nvidia,
                MKDEV(NV_MAJOR_DEVICE_NUMBER, i), &nv_linux_devices[i].dev->dev, name);
        }
    }
#endif

    nv_printf(NV_DBG_INFO, "NVRM: major number %d\n", nv_major);

    /* instantiate tasklets */
    for (i = 0; i < NV_MAX_DEVICES; i++)
    {
        /*
         * We keep one tasklet per card to avoid latency issues with more
         * than one device; no two instances of a single tasklet are ever
         * executed concurrently.
         */
        NV_ATOMIC_SET(nv_linux_devices[i].tasklet.count, 1);
    }

    // init the nvidia control device
    {
        nv_state_t *nv_ctl = NV_STATE_PTR(&nv_ctl_device);
        nv_ctl->os_state = (void *) &nv_ctl_device;
        nv_lock_init_locks(nv_ctl);
    }

#if defined(NV_PM_SUPPORT_APM)
    for (i = 0; i < num_nv_devices; i++)
    {
        apm_nv_dev[i] = pm_register(PM_PCI_DEV, PM_SYS_VGA, nv_kern_apm_event);
    }
#endif

    // Init the resource manager
    if (!rm_init_rm())
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: rm_init_rm() failed\n");
        rc = -EIO;
        goto failed;
    }

    // load our local registry entries into the registry
    {
        extern nv_parm_t nv_parms[];
        rm_load_registry(nv_parms);
    }

    /* create /proc/driver/nvidia */
    nvos_proc_create();

#if defined(DEBUG)
    inter_module_register("nv_linux_devices", THIS_MODULE, nv_linux_devices);
#endif

    /* Register ioctl conversions for 32 bit clients */
    rm_register_ioctl_conversions();

#ifdef NV_SWIOTLB
    // may need a better test than this eventually
    if (swiotlb)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: WARNING:  You are probably using the kernel's swiotlb interface.\n");
        nv_printf(NV_DBG_ERRORS, "NVRM: Be very careful with this interface, as it is easy\n");
        nv_printf(NV_DBG_ERRORS, "NVRM: to exhaust this interface's memory buffer, at which\n");
        nv_printf(NV_DBG_ERRORS, "NVRM: point it panics the kernel.  Please increase the size\n");
        nv_printf(NV_DBG_ERRORS, "NVRM: of this buffer by specifying a larger buffer size with\n");
        nv_printf(NV_DBG_ERRORS, "NVRM: the swiotlb kernel option, eg: \"swiotlb=16384\"\n");
        nv_swiotlb = 1;
    }
#endif

    return 0;

failed:
#ifdef CONFIG_DEVFS_FS
    NV_DEVFS_REMOVE_CONTROL();
    for (i = 0; i < num_nv_devices; i++)
        NV_DEVFS_REMOVE_DEVICE(i);
#endif

    if (NV_UNREGISTER_CHRDEV(nv_major, "nvidia") < 0)
        nv_printf(NV_DBG_ERRORS, "NVRM: unregister nv chrdev failed\n");

#ifdef NV_CLASS_SIMPLE_PRESENT
    class_simple_device_remove(MKDEV(NV_MAJOR_DEVICE_NUMBER, 255));
    for (i = 0; i < num_nv_devices; i++)
        class_simple_device_remove(MKDEV(NV_MAJOR_DEVICE_NUMBER, i));
    class_simple_destroy(class_nvidia);
#endif

    for (i = 0; i < num_nv_devices; i++)
    {
        if (nv_linux_devices[i].dev)
        {
            nv_state_t *nv = NV_STATE_PTR(&nv_linux_devices[i]);
            release_mem_region(nv->bar.regs.address, nv->bar.regs.size);
        }
    }

    pci_unregister_driver(&nv_pci_driver);
    return rc;
}

static void __exit nvidia_exit_module(void)
{
    int i;
    nv_linux_state_t *nvl, *max_devices;

    nv_printf(NV_DBG_INFO, "NVRM: nvidia_exit_module\n");

#ifdef CONFIG_DEVFS_FS
    NV_DEVFS_REMOVE_CONTROL();
    for (i = 0; i < num_nv_devices; i++)
        NV_DEVFS_REMOVE_DEVICE(i);
#endif

    if (NV_UNREGISTER_CHRDEV(nv_major, "nvidia") < 0)
        nv_printf(NV_DBG_ERRORS, "NVRM: unregister nv chrdev failed\n");

#ifdef NV_CLASS_SIMPLE_PRESENT
    class_simple_device_remove(MKDEV(NV_MAJOR_DEVICE_NUMBER, 255));
    for (i = 0; i < num_nv_devices; i++)
        class_simple_device_remove(MKDEV(NV_MAJOR_DEVICE_NUMBER, i));
    class_simple_destroy(class_nvidia);
#endif

    for (i = 0; i < num_nv_devices; i++)
    {
        if (nv_linux_devices[i].dev)
        {
            nv_state_t *nv = NV_STATE_PTR(&nv_linux_devices[i]);
            release_mem_region(nv->bar.regs.address, nv->bar.regs.size);
        }
    }

    pci_unregister_driver(&nv_pci_driver);

    /* remove /proc/driver/nvidia */
    nvos_proc_remove();

#if defined(DEBUG)
    inter_module_unregister("nv_linux_devices");
#endif

#if defined(NV_PM_SUPPORT_APM)
    for (i = 0; i < num_nv_devices; i++)
    {
        pm_unregister(apm_nv_dev[i]);
    }
#endif

    // Shutdown the resource manager
    rm_shutdown_rm();

    /*
     * Make sure we freed up all the mappings. The kernel should
     * do this automatically before calling close.
     */
    max_devices = nv_linux_devices + NV_MAX_DEVICES;
    for (nvl = nv_linux_devices; nvl < max_devices; nvl++)
    {
        nv_state_t *nv = NV_STATE_PTR(nvl);

        if (nvl->alloc_queue)
        {
            nv_alloc_t *at = nvl->alloc_queue;
            while (at)
            {
                NV_PRINT_AT(at);
                at = at->next;
            }
        }

        if (nv->flags & NV_FLAG_MAP_REGS_EARLY)
        {
            NV_ASSERT("regs not mapped when they should be!",
                      nv->bar.regs.map != NULL);
            os_unmap_kernel_space(nv->bar.regs.map, nv->bar.regs.size);
            nv->bar.regs.map = NULL;
            nv->bar.regs.map_u = (nv_phwreg_t) NULL;
        }
    }

    /* Unregister ioctl conversions for 32 bit clients */
    rm_unregister_ioctl_conversions();

#if defined(NV_BUILD_NV_PAT_SUPPORT)
    if (pat_enabled)
    {
        __nv_disable_pat_support();
    }
#endif

#if NV_ENABLE_MEM_TRACKING
    nv_list_mem(vm_list);
    nv_list_mem(km_list);
    nv_list_mem(fp_list);
    nv_printf(NV_DBG_ERRORS, "NVRM: final mem usage: vm 0x%x km 0x%x fp 0x%x\n",
        vm_usage, km_usage, fp_usage);
#endif
}

module_init(nvidia_init_module);
module_exit(nvidia_exit_module);


/* this is only called when the vmas are duplicated.
 * this appears to only happen when the process is cloned to create
 * a new process, and not when the process is threaded.
 *
 * increment the usage count for the physical pages, so when this
 * clone unmaps the mappings, the pages are not deallocated under
 * the original process.
 */
void
nv_kern_vma_open(struct vm_area_struct *vma)
{
    nv_printf(NV_DBG_MEMINFO, "NVRM: vma_open for 0x%x - 0x%x, offset 0x%x\n",
        vma->vm_start, vma->vm_end, NV_VMA_OFFSET(vma));

    if (NV_VMA_PRIVATE(vma))
    {
        nv_alloc_t *at = (nv_alloc_t *) NV_VMA_PRIVATE(vma);
        NV_ATOMIC_INC(at->usage_count);

        nv_printf(NV_DBG_MEMINFO,
            "NVRM:   at 0x%x, usage count %d, page_table 0x%x\n",
            at, at->usage_count, at->page_table);

        nvos_list_page_count(at->page_table, at->num_pages);
    }
}


void
nv_kern_vma_release(struct vm_area_struct *vma)
{
    nv_printf(NV_DBG_MEMINFO,
        "NVRM: vma_release for 0x%x - 0x%x, offset 0x%x\n",
        vma->vm_start, vma->vm_end, NV_VMA_OFFSET(vma));

    if (NV_VMA_PRIVATE(vma))
    {
        nv_alloc_t *at = (nv_alloc_t *) NV_VMA_PRIVATE(vma);

        NV_ATOMIC_DEC(at->usage_count);

        nv_printf(NV_DBG_MEMINFO,
            "NVRM:  at 0x%x, usage count %d, page_table 0x%x\n",
            at, at->usage_count, at->page_table);

        nvos_list_page_count(at->page_table, at->num_pages);

        // if usage_count is down to 0, the kernel virtual mapping was freed
        // but the underlying physical pages were not, due to the reserved bit
        // being set. We need to clear the reserved bit, then munmap will
        // zap the pages and free the physical pages.
        if (NV_ATOMIC_READ(at->usage_count) == 0)
        {
            if (at->page_table)
                nvos_unlock_pages(at);
            nvos_free_alloc(at);
            NV_VMA_PRIVATE(vma) = NULL;
        }
    }
}


/* at this point, this code just plain won't work with 2.2 kernels.
 * additionally, only ia64 & the 460GX need a nopage handler, and 2.2 doesn't
 * work on ia64 anyways. It's expected that at some point other agp chipsets
 * will work similar to the 460GX (AGP 3.0 spec), so pre-emptively make sure
 * this works on our standard ia32 driver.
 */

/* AGP allocations under the 460GX are not mapped to the aperture
 * addresses by the CPU.  This nopage handler will fault on CPU
 * accesses to AGP memory and map the address to the correct page.
 */
struct page *nv_kern_vma_nopage(
    struct vm_area_struct *vma,
    unsigned long address,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 1))
    int *type
#else
    int write_access
#endif
)
{
#if defined(NVCPU_IA64) && (LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 9))
    nv_alloc_t *at, *tmp;
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    struct page *page_ptr;
    int rm_status, index;

    at = NV_VMA_PRIVATE(vma);
    if (at == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: nopage handler called without an at: "
                  "vm_start 0x%x, at 0x%x\n", vma->vm_start, at);
        return NOPAGE_SIGBUS;
    }

    // let's verify this 'at' is valid
    // I can imagine cases where something went wrong, the 'at' and underlying
    // pages were freed, but the virtual mapping still exists and this 'at'
    // pointer is potentially pointing to freed memory. Let's make sure we can
    // still find the 'at' in our alloc_queue.
    nvl = NVL_FROM_FILEP(vma->vm_file);
    if (nvl == NULL)
        return NOPAGE_SIGBUS;

    nv = (nv_state_t *) nvl;

    rm_status = RM_ERROR;
    tmp = nvl->alloc_queue;
    while (tmp)
    {
        if (tmp == at)
        {
            rm_status = RM_OK;
            break;
        }
        tmp = tmp->next;
    }

    if (rm_status != RM_OK)
    {
        // we didn't find the 'at' (and haven't dereferenced it yet).
        // let's bail before something bad happens, but first print an
        // error message and NULL the pointer out so we don't come this
        // far again
        nv_printf(NV_DBG_ERRORS, "NVRM: nopage handler called on a freed"
                  "address: vm_start 0x%x, at 0x%x\n", vma->vm_start, at);
        NV_VMA_PRIVATE(vma) = NULL;
        return NOPAGE_SIGBUS;
    }

    rm_status = KernMapAGPNopage((void *)address, vma, at->priv_data, 
                                 (void **)&page_ptr);
    if (rm_status)
        return NOPAGE_SIGBUS;

    // get the index of this page into the allocation
    index = (address - vma->vm_start)>>PAGE_SHIFT;

    // save that index into our page list (make sure it doesn't already exist)
    if (at->page_table[index].phys_addr)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: page slot already filled in nopage handler!\n");
        os_dbg_breakpoint();
    }

    at->page_table[index].phys_addr = (page_to_pfn(page_ptr) << PAGE_SHIFT);
    at->page_table[index].dma_addr  = (page_to_pfn(page_ptr) << PAGE_SHIFT);
    at->page_table[index].virt_addr = (unsigned long) __va(page_to_pfn(page_ptr) << PAGE_SHIFT);

    return page_ptr;
#endif
    return NOPAGE_SIGBUS;
}

struct vm_operations_struct nv_vm_ops = {
    .open   = nv_kern_vma_open,
    .close  = nv_kern_vma_release,  /* "close" */
    .nopage = nv_kern_vma_nopage,
};

static nv_file_private_t *
nv_alloc_file_private(void)
{
    nv_file_private_t *nvfp;

    NV_VMALLOC(nvfp, sizeof(nv_file_private_t));
    if (!nvfp)
        return NULL;

    memset(nvfp, 0, sizeof(nv_file_private_t));

    // initialize this file's event queue
    init_waitqueue_head(&nvfp->waitqueue);

    nv_init_lock(nvfp->fp_lock);

    NV_VMALLOC(nvfp->event_fifo, sizeof(nv_event_t) * NV_EVENT_FIFO_SIZE);
    if (nvfp->event_fifo == NULL)
    {
        NV_VFREE(nvfp, sizeof(nv_file_private_t));
        return NULL;
    }

    return nvfp;
}

static void
nv_free_file_private(nv_file_private_t *nvfp)
{
    if (nvfp == NULL)
        return;

    NV_VFREE(nvfp->event_fifo, sizeof(nv_event_t) * NV_EVENT_FIFO_SIZE);
    NV_VFREE(nvfp, sizeof(nv_file_private_t));
}


/*
** nv_kern_open
**
** nv driver open entry point.  Sessions are created here.
*/
int nv_kern_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv = NULL;
    nv_linux_state_t *nvl = NULL;
    int devnum;
    int rc = 0, status;

    nv_printf(NV_DBG_INFO, "NVRM: nv_kern_open...\n");

    FILE_PRIVATE(file) = nv_alloc_file_private();
    if (FILE_PRIVATE(file) == NULL)
        return -ENOMEM;

    /* for control device, just jump to its open routine */
    /* after setting up the private data */
    if (NV_IS_CONTROL_DEVICE(inode))
        return nv_kern_ctl_open(inode, file);

    /* what device are we talking about? */
    devnum = NV_DEVICE_NUMBER(inode);
    if (devnum >= NV_MAX_DEVICES)
    {
        rc = -ENODEV;
        goto failed;
    }

    nvl = &nv_linux_devices[devnum];
    nv = NV_STATE_PTR(nvl);

    nv_printf(NV_DBG_INFO, "NVRM: nv_kern_open on device %d\n", devnum);
    nv_down(nvl->ldata_lock);

    NV_CHECK_PCI_CONFIG(nv);

    NVL_FROM_FILEP(file) = nvl;

    /*
     * map the memory and allocate isr on first open
     */

    if ( ! (nv->flags & NV_FLAG_OPEN))
    {
        if (nv->device_id == 0)
        {
            nv_printf(NV_DBG_ERRORS, "NVRM: open of nonexistent device %d\n",
                devnum);
            rc = -ENXIO;
            goto failed;
        }

        status = request_irq(nv->interrupt_line, nv_kern_isr,
                             SA_INTERRUPT | SA_SHIRQ, "nvidia",
                             (void *) nvl);
        if (status != 0)
        {
            if ( nv->interrupt_line && (status == -EBUSY) )
            {
                nv_printf(NV_DBG_ERRORS,
                    "NVRM: Tried to get irq %d, but another driver",
                    (unsigned int) nv->interrupt_line);
                nv_printf(NV_DBG_ERRORS, "NVRM: has it and is not sharing it.\n");
                nv_printf(NV_DBG_ERRORS, "NVRM: you may want to verify that an audio driver");
                nv_printf(NV_DBG_ERRORS, " isn't using the irq\n");
            }
            nv_printf(NV_DBG_ERRORS, "NVRM: isr request failed 0x%x\n", status);
            rc = -EIO;
            goto failed;
        }

        if ( ! rm_init_adapter(nv))
        {
            free_irq(nv->interrupt_line, (void *) nvl);
            nv_printf(NV_DBG_ERRORS, "NVRM: rm_init_adapter failed\n");
            rc = -EIO;
            goto failed;
        }

        nvl->tasklet.func = nv_kern_isr_bh;
        nvl->tasklet.data = (unsigned long) nv;
        tasklet_enable(&nvl->tasklet);

        nv->flags |= NV_FLAG_OPEN;
    }

    NV_ATOMIC_INC(nvl->usage_count);

 failed:
    nv_up(nvl->ldata_lock);

    if ((rc) && FILE_PRIVATE(file))
    {
        nv_free_file_private(FILE_PRIVATE(file));
        FILE_PRIVATE(file) = NULL;
    }

    return rc;
}


/*
** nv_kern_close
**
** Master driver close entry point.
*/

int nv_kern_close(
    struct inode *inode,
    struct file *file
)
{
    nv_linux_state_t *nvl = NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);

    NV_CHECK_PCI_CONFIG(nv);

    /* for control device, just jump to its open routine */
    /* after setting up the private data */
    if (NV_IS_CONTROL_DEVICE(inode))
        return nv_kern_ctl_close(inode, file);

    nv_printf(NV_DBG_INFO, "NVRM: nv_kern_close on device %d\n",
        NV_DEVICE_NUMBER(inode));

    rm_free_unused_clients(nv, current->pid, (void *) file);

    nv_down(nvl->ldata_lock);
    if (NV_ATOMIC_DEC_AND_TEST(nvl->usage_count))
    {
        /*
         * The usage count for this device has dropped to zero, it can be shut
         * down safely; disable its interrupts.
         */
        rm_disable_adapter(nv);

        /*
         * Disable this device's tasklet to make sure that no bottom half will
         * run with undefined device state.
         */
        tasklet_disable(&nvl->tasklet);

        /*
         * Free the IRQ, which may block until all pending interrupt processing
         * has completed.
         */
        free_irq(nv->interrupt_line, (void *) nvl);

        rm_shutdown_adapter(nv);

        /*
         * Make sure we have freed up all the mappings. The kernel
         * should do this automagically before calling close
         */
        if (nvl->alloc_queue)
        {
            nv_alloc_t *at = nvl->alloc_queue;
            while (at)
            {
                NV_PRINT_AT(at);
                at = at->next;
            }
        }

        /* leave INIT flag alone so we don't reinit every time */
        nv->flags &= ~NV_FLAG_OPEN;
    }
    nv_up(nvl->ldata_lock);

    if (FILE_PRIVATE(file))
    {
        nv_free_file_private(FILE_PRIVATE(file));
        FILE_PRIVATE(file) = NULL;
    }

    return 0;
}

int nv_kern_mmap(
    struct file  *file,
    struct vm_area_struct *vma
)
{
    int pages;
    nv_alloc_t *at;
    nv_linux_state_t *nvl = NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);

    nv_printf(NV_DBG_INFO, "NVRM: mmap([0x%p-0x%p] off=0x%lx)\n",
        vma->vm_start,
        vma->vm_end,
        NV_VMA_OFFSET(vma));

    NV_CHECK_PCI_CONFIG(nv);

    // be a bit paranoid for now
    if ((NV_MASK_OFFSET(NV_VMA_OFFSET(vma))) ||
        (NV_MASK_OFFSET(vma->vm_start)) ||
        (NV_MASK_OFFSET(vma->vm_end)))
    {
        return -ENXIO;
    }

    pages = (vma->vm_end - vma->vm_start) >> PAGE_SHIFT;

    // we have our own version to keep the module count right
    vma->vm_ops = &nv_vm_ops;

    /*
     * figure out the range and map it in
     */


    /* NV reg space */
    if (IS_REG_OFFSET(nv, NV_VMA_OFFSET(vma), vma->vm_end - vma->vm_start))
    {
        /* truncate to size of registers */
        if (pages > nv->regs->size / PAGE_SIZE)
            pages = nv->regs->size / PAGE_SIZE;

        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        if (NV_REMAP_PAGE_RANGE(vma->vm_start,
                             NV_VMA_OFFSET(vma),
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        /* mark it as IO so that we don't dump it on core dump */
        vma->vm_flags |= VM_IO;
    }

    /* NV fb space */
    else if (IS_FB_OFFSET(nv, NV_VMA_OFFSET(vma), vma->vm_end - vma->vm_start))
    {

        /* truncate to size of framebuffer */
        if (pages > nv->fb->size / PAGE_SIZE)
            pages = nv->fb->size / PAGE_SIZE;

        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        if (NV_REMAP_PAGE_RANGE(vma->vm_start,
                             NV_VMA_OFFSET(vma),
                             vma->vm_end - vma->vm_start,
                             vma->vm_page_prot))
            return -EAGAIN;

        // mark it as IO so that we don't dump it on core dump
        vma->vm_flags |= VM_IO;
    }

    /* AGP allocator */
    else if (IS_AGP_OFFSET(nv, NV_VMA_OFFSET(vma), vma->vm_end - vma->vm_start))
    {
        nv_down(nvl->at_lock);
        at = nvl_find_alloc(nvl, NV_VMA_OFFSET(vma), NV_ALLOC_TYPE_AGP);

        if (at == NULL)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: couldn't find pre-allocated agp memory!\n");
            nv_up(nvl->at_lock);
            return -EAGAIN;
        }

        if (at->num_pages != pages)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: pre-allocated agp memory has wrong number of pages!\n");
            nv_up(nvl->at_lock);
            return -EAGAIN;
        }

        NV_VMA_PRIVATE(vma) = at;
        NV_ATOMIC_INC(at->usage_count);
        nv_up(nvl->at_lock);

        if (NV_OSAGP_ENABLED(nv))
        {
#if !defined(NVCPU_IA64) || (LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9))
            KernMapAGPPages(vma, at->priv_data);
#else
            /* Note: on IA64 the AGP chipset is cache coherent, so we
             * leave the AGP allocation mapped cached. */
#endif
        }
        else
        {
            rm_map_agp_pages(nv, (void **) &vma, at->class, at->priv_data);
        }
        nvos_list_page_count(at->page_table, at->num_pages);

        /* prevent the swapper from swapping it out */
        /* mark the memory i/o so the buffers aren't dumped on core dumps */
        vma->vm_flags |= (VM_LOCKED | VM_IO);
    }

    /* Magic allocator */
    else // if (NV_VMA_OFFSET(vma) == NV_MMAP_ALLOCATION_OFFSET)
    {
        unsigned long page = 0, pos, start;
        int i = 0;

        nv_down(nvl->at_lock);
        at = nvl_find_alloc(nvl, NV_VMA_OFFSET(vma), NV_ALLOC_TYPE_PCI);

        if (at == NULL)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: couldn't find pre-allocated memory!\n");
            nv_up(nvl->at_lock);
            return -EAGAIN;
        }

        if (at->num_pages != pages)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: pre-allocated sys memory has wrong number of pages!\n");
            nv_up(nvl->at_lock);
            return -EAGAIN;
        }

        // allow setting or refusal of specific caching types
        switch (NV_ALLOC_MAPPING(at->flags))
        {
            case NV_MEMORY_DEFAULT:
            case NV_MEMORY_WRITEBACK:
                break;
            case NV_MEMORY_UNCACHED:
                vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
                break;
            case NV_MEMORY_WRITECOMBINED:
#if defined(NV_BUILD_NV_PAT_SUPPORT)
                if (pat_enabled)
                {
                    vma->vm_page_prot = pgprot_writecombined(vma->vm_page_prot);
                    break;
                }
#endif
            case NV_MEMORY_WRITETHRU:
            case NV_MEMORY_WRITEPROTECT:
            default:
                nv_printf(NV_DBG_INFO,
                    "NVRM: memory caching type 0x%x not supported!\n",
                    NV_ALLOC_MAPPING(at->flags));
                nv_up(nvl->at_lock);
                return -EAGAIN;
        }

        NV_VMA_PRIVATE(vma) = at;
        NV_ATOMIC_INC(at->usage_count);
        nv_up(nvl->at_lock);

        nv_printf(NV_DBG_INFO, "NVRM: remapping %d system pages for at 0x%x\n",
            pages, at);
        start = vma->vm_start;
        while (pages--)
        {
            page = (unsigned long) at->page_table[i++].phys_addr;
            if (NV_REMAP_PAGE_RANGE(start, page, PAGE_SIZE, vma->vm_page_prot))
                return -EAGAIN;
            start += PAGE_SIZE;
            pos += PAGE_SIZE;
        }
        nvos_list_page_count(at->page_table, at->num_pages);

        /* prevent the swapper from swapping it out */
        /* mark the memory i/o so the buffers aren't dumped on core dumps */
        vma->vm_flags |= (VM_IO | VM_LOCKED);
    }

    vma->vm_file = file;

    return 0;
}


unsigned int nv_kern_poll(
    struct file *file,
    poll_table *wait
)
{
    unsigned int mask = 0;
    nv_file_private_t *nvfp;
    nv_linux_state_t *nvl;
    unsigned long eflags;

    nvl = NVL_FROM_FILEP(file);

    if (NV_STATE_PTR(nvl)->device_number == NV_CONTROL_DEVICE_NUMBER)
        return nv_kern_ctl_poll (file, wait);

    nvfp = NV_GET_NVFP(file);

    if ( !(file->f_flags & O_NONBLOCK))
    {
        // add us to the list
        poll_wait(file, &nvfp->waitqueue, wait);
    }

    nv_lock_irq(nvfp->fp_lock, eflags);

    // wake the user on any event
    if (nvfp->num_events)
    {
        nv_printf(NV_DBG_EVENTINFO, "NVRM: Hey, an event occured!\n");
        // trigger the client, when they grab the event, 
        // we'll decrement the event count
        mask |= (POLLPRI|POLLIN);
    }

    nv_unlock_irq(nvfp->fp_lock, eflags);

    return mask;
}

//
// nv_kern_ioctl
//
// nv driver ioctl entry point.
//

/*
 * some ioctl's can only be done on actual device, others only on the control device
 */
#define CTL_DEVICE_ONLY(nv) { if ( ! ((nv)->flags & NV_FLAG_CONTROL)) { status = -EINVAL; goto done; } }

#define ACTUAL_DEVICE_ONLY(nv) { if ((nv)->flags & NV_FLAG_CONTROL) { status = -EINVAL; goto done; } }


/* todo:
   need ioctl to raise a thread priority that is not superuser
       set its priority to SCHED_FIFO which is simple
       priority scheduling w/ disabled timeslicing
*/

int nv_kern_ioctl(
    struct inode *inode,
    struct file *file,
    unsigned int cmd,
    unsigned long i_arg)
{
    int status = 0;
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    void *arg = (void *) i_arg;
    void *arg_copy;
    int arg_size;

    nvl = NVL_FROM_FILEP(file);
    nv = NV_STATE_PTR(nvl);

    nv_printf(NV_DBG_INFO, "NVRM: ioctl(0x%x, 0x%x, 0x%x)\n",
        _IOC_NR(cmd), (unsigned int) i_arg, _IOC_SIZE(cmd));

    NV_CHECK_PCI_CONFIG(nv);

    arg_size = _IOC_SIZE(cmd);
    NV_KMALLOC(arg_copy, arg_size);
    if (arg_copy == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate ioctl memory\n");
        return -ENOMEM;
    }

    if (copy_from_user(arg_copy, arg, arg_size))
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to copy ioctl data\n");
        NV_KFREE(arg_copy, arg_size);
        return -ENOMEM;
    }

    switch (_IOC_NR(cmd))
    {
        /* pass out info about the card */
        case NV_ESC_CARD_INFO:
        {
            nv_ioctl_card_info_t *ci;
            nv_linux_state_t *tnvl;
            nv_ioctl_rm_api_version_t *rm_api;
            int i;

            CTL_DEVICE_ONLY(nv);

            /* the first element of card info passed from the client will have
             * the rm_api_version_magic value to show that the client is new
             * enough to support versioning. If the client is too old to 
             * support versioning, our mmap interfaces are probably different
             * enough to cause serious damage.
             * just copy in the one dword to check.
             */
            rm_api = arg_copy;
            if ((rm_api->magic   != NV_RM_API_VERSION_MAGIC_REQ) ||
                (rm_api->version != NV_RM_API_VERSION))
            {
                if (rm_api->magic != NV_RM_API_VERSION_MAGIC_REQ)
                {
                    nv_printf(NV_DBG_ERRORS, 
                        "NVRM: client does not support versioning!!\n");
                } else
                if (rm_api->version != NV_RM_API_VERSION)
                {
                    nv_printf(NV_DBG_ERRORS, 
                        "NVRM: client supports wrong rm api version!!\n");
                }
                nv_printf(NV_DBG_ERRORS,
                    "NVRM:    aborting to avoid catastrophe!\n");
                rm_api->magic   = NV_RM_API_VERSION_MAGIC_REP;
                rm_api->version = NV_RM_API_VERSION;
                rm_api->major   = NV_MAJOR_VERSION;
                rm_api->minor   = NV_MINOR_VERSION;
                rm_api->patch   = NV_PATCHLEVEL;
                status = -EINVAL;
                break;
            }

            ci = arg_copy;
            memset(ci, 0, arg_size);
            for (i = 0, tnvl = nv_linux_devices; tnvl < nv_linux_devices + NV_MAX_DEVICES; tnvl++, i++)
            {
                nv_state_t *tnv = NV_STATE_PTR(tnvl);
                if (tnv->device_id)
                {
                    ci->flags = NV_IOCTL_CARD_INFO_FLAG_PRESENT;
                    ci->bus = tnv->bus;
                    ci->slot = tnv->slot;
                    ci->vendor_id = tnv->vendor_id;
                    ci->device_id = tnv->device_id;
                    ci->interrupt_line = tnv->interrupt_line;
                    ci->reg_address = tnv->bar.regs.address;
                    ci->reg_size = tnv->bar.regs.size;
                    ci->fb_address = tnv->bar.fb.address;
                    ci->fb_size = tnv->bar.fb.size;
                    ci++;
                }
            }
            break;
        }

        /* set a card to be posted */
        case NV_ESC_POST_VBIOS:
        {
            nv_ioctl_post_vbios_t *params = arg_copy;

            CTL_DEVICE_ONLY(nv);

            status = nvos_post_vbios(params);

            break;
        }

        /* get the sim environment info for this setup */
        case NV_ESC_SIM_ENV:
        {
            nv_ioctl_sim_env_t *simenv = arg_copy;

            CTL_DEVICE_ONLY(nv);

            simenv->sim_env = nv->sim_env;

            break;
        }

        case NV_ESC_RM_API_VERSION:
        {
            nv_ioctl_rm_api_version_t *rm_api = arg_copy;

            CTL_DEVICE_ONLY(nv);

            rm_api->version = NV_RM_API_VERSION;
            rm_api->major   = NV_MAJOR_VERSION;
            rm_api->minor   = NV_MINOR_VERSION;
            rm_api->patch   = NV_PATCHLEVEL;

            break;
        }


        default:
            status = rm_ioctl(nv, file, _IOC_NR(cmd), arg_copy) ? 0 : -EINVAL;
            break;
    }

 done:
    if (RM_OK != copy_to_user(arg, arg_copy, arg_size))
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to copyout ioctl data\n");
    NV_KFREE(arg_copy, arg_size);
    return status;
}

/*
 * driver receives an interrupt
 *    if someone waiting, then hand it off.
 */
irqreturn_t nv_kern_isr(
    int   irq,
    void *arg,
    struct pt_regs *regs
)
{
    nv_linux_state_t *nvl = (void *) arg;
    nv_state_t *nv = NV_STATE_PTR(nvl);
    U032 need_to_run_bottom_half = 0;

    NV_CHECK_PCI_CONFIG(nv);
    rm_isr(nv->device_number, &need_to_run_bottom_half);
    if (need_to_run_bottom_half)
    {
        tasklet_schedule(&nvl->tasklet);
    }

    return IRQ_HANDLED;
}

void nv_kern_isr_bh(
    unsigned long data
)
{
    nv_state_t *nv = (nv_state_t *) data;
    /*
     * XXX: This level of indirection is necessary to work around
     * problems with Linux kernels using a non-standard calling
     * convention, i.e. Arjan van de Ven's/RedHat's 2.6.0 kernels.
     */
    NV_CHECK_PCI_CONFIG(nv);
    rm_isr_bh(nv->pdev);
}

void nv_kern_rc_timer(
    unsigned long data
)
{
    nv_linux_state_t *nvl = (nv_linux_state_t *) data;

    // nv_printf(NV_DBG_INFO, "NVRM: rc timer\n");

    NV_CHECK_PCI_CONFIG((nv_state_t *) data);
    rm_run_rc_callback((nv_state_t *) data);
    mod_timer(&nvl->rc_timer, jiffies + HZ);  /* set another timeout in 1 second */
}

#if defined(NV_PM_SUPPORT_APM)
/* kernel calls us with a power management event */
int
nv_kern_apm_event(
    struct pm_dev *dev,
    pm_request_t rqst,
    void *data
)
{
    nv_state_t *nv;
    nv_linux_state_t *lnv;
    int devnum;
    int status = RM_OK;

    nv_printf(NV_DBG_INFO, "NVRM: nv_kern_apm_event: %d (0x%p)\n", rqst, data);

    for (devnum = 0; devnum < num_nv_devices; devnum++)
    {
        if (apm_nv_dev[devnum] == dev)
        {
            break;
        }
    }

    if (devnum == num_nv_devices)
    {
        nv_printf(NV_DBG_WARNINGS, "NVRM: APM: invalid device!\n");
        return 1;
    }

    lnv = &nv_linux_devices[devnum];
    nv = NV_STATE_PTR(lnv);

    if (nv->pdev == NULL)
    {
        nv_printf(NV_DBG_WARNINGS, "NVRM: APM: device not initialized!\n");
        return 1;
    }

    NV_CHECK_PCI_CONFIG(NV_STATE_PTR(lnv));

    switch (rqst)
    {
#if defined(CONFIG_APM) || defined(CONFIG_APM_MODULE)
        case PM_RESUME:
            nv_printf(NV_DBG_INFO, "NVRM: APM: received resume event\n");
            status = rm_power_management(nv, 0, NV_PM_APM_RESUME);
            break;

        case PM_SUSPEND:
            nv_printf(NV_DBG_INFO, "NVRM: APM: received suspend event\n");
            status = rm_power_management(nv, 0, NV_PM_APM_SUSPEND);
            break;
#endif
        default:
            nv_printf(NV_DBG_WARNINGS, "NVRM: APM: unsupported event: %d\n", rqst);
            return 1;
    }

    if (status != RM_OK)
        nv_printf(NV_DBG_ERRORS, "NVRM: APM: failed event: %d\n", rqst);

    return status;
}
#endif

/*
** nv_kern_ctl_open
**
** nv control driver open entry point.  Sessions are created here.
*/
int nv_kern_ctl_open(
    struct inode *inode,
    struct file *file
)
{
    nv_state_t *nv;
    nv_linux_state_t *nvl;
    int rc = 0;

    nvl = &nv_ctl_device;
    nv = (nv_state_t *) nvl;

    nv_printf(NV_DBG_INFO, "NVRM: nv_kern_ctl_open\n");

    nv_down(nvl->ldata_lock);

    nv->device_number = NV_CONTROL_DEVICE_NUMBER;

    /* save the nv away in file->private_data */
    NVL_FROM_FILEP(file) = nvl;

    if (NV_ATOMIC_READ(nvl->usage_count) == 0)
    {
        init_waitqueue_head(&nv_ctl_waitqueue);
    }

    nv->flags |= NV_FLAG_OPEN + NV_FLAG_CONTROL;

    /* turn off the hotkey occurred bit */
    nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;

    NV_ATOMIC_INC(nvl->usage_count);
    nv_up(nvl->ldata_lock);

    return rc;
}


/*
** nv_kern_ctl_close
*/
int nv_kern_ctl_close(
    struct inode *inode,
    struct file *file
)
{
    nv_linux_state_t *nvl =  NVL_FROM_FILEP(file);
    nv_state_t *nv = NV_STATE_PTR(nvl);

    nv_printf(NV_DBG_INFO, "NVRM: nv_kern_ctl_close\n");

    nv_down(nvl->ldata_lock);
    if (NV_ATOMIC_DEC_AND_TEST(nvl->usage_count))
    {
        nv->flags = 0;
    }
    nv_up(nvl->ldata_lock);

    rm_free_unused_clients(nv, current->pid, (void *) file);

    if (FILE_PRIVATE(file))
    {
        nv_free_file_private(FILE_PRIVATE(file));
        FILE_PRIVATE(file) = NULL;
    }

    return 0;
}


/*
 * nv_kern_ctl_poll() - add the process to the wait queue
 */

unsigned int nv_kern_ctl_poll(
    struct file *file,
    poll_table *wait
)
{
    nv_linux_state_t *nvl;
    nv_state_t *nv;
    unsigned int ret = 0;

    nvl = NVL_FROM_FILEP(file);
    nv = NV_STATE_PTR(nvl);

    if ( !(file->f_flags & O_NONBLOCK) )
    {
        poll_wait(file, &nv_ctl_waitqueue, wait);
    }

    nv_lock_rm(nv);
    if (nv->flags & NV_FLAG_HOTKEY_OCCURRED)
    {
        nv_printf(NV_DBG_EVENTINFO, "NVRM: a hotkey event has occurred\n");
        nv->flags &= ~NV_FLAG_HOTKEY_OCCURRED;
        ret = POLLIN | POLLRDNORM;
    }
    nv_unlock_rm(nv);

    return ret;
}




/*
 * nv_set_hotkey_occurred_flag() - set the hotkey flag and wake up anybody
 * waiting on the wait queue
 */

void NV_API_CALL nv_set_hotkey_occurred_flag(void)
{
    nv_state_t *nv = NV_STATE_PTR(&nv_ctl_device);

    nv_printf(NV_DBG_EVENTINFO, "NVRM: setting the hotkey occurred flag!\n");

    nv_lock_rm(nv);
    nv_ctl_device.nv_state.flags |= NV_FLAG_HOTKEY_OCCURRED;
    nv_unlock_rm(nv);

    wake_up_interruptible(&nv_ctl_waitqueue);
}

int nv_kern_read_cardinfo(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    struct pci_dev *dev;
    char *type, *fmt, tmpstr[NV_DEVICE_NAME_LENGTH];
    int len = 0, status;
    U032 vbios_rev1, vbios_rev2, vbios_rev3, vbios_rev4, vbios_rev5;

    nv_state_t *nv;
    nv = (nv_state_t *) data;

    dev = nv_find_pci_dev(nv);
    if (!dev)
        return 0;
    
    if (rm_get_device_name(nv, dev->device, NV_DEVICE_NAME_LENGTH,
                           tmpstr) != RM_OK) {
        strcpy (tmpstr, "Unknown");
    }
    
    len += sprintf(page+len, "Model: \t\t %s\n", tmpstr);
    len += sprintf(page+len, "IRQ:   \t\t %d\n", nv->interrupt_line);

    status = rm_get_vbios_version(nv, &vbios_rev1, &vbios_rev2,
                                  &vbios_rev3, &vbios_rev4, &vbios_rev5);

    if (status < 0) {
        /* before rm_init_adapter */
        len += sprintf(page+len, "Video BIOS: \t ??.??.??.??.??\n");
    } else {
        fmt = "Video BIOS: \t %02x.%02x.%02x.%02x.%02x\n";
        len += sprintf(page+len, fmt, vbios_rev1, vbios_rev2, vbios_rev3,
                                                  vbios_rev4, vbios_rev5);
    }

    type = nvos_find_agp_capability(dev) ? "AGP" : "PCI";
    len += sprintf(page+len, "Card Type: \t %s\n", type);

    return len;
}

int nv_kern_read_version(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    int len = 0;
    
    len += sprintf(page+len, "NVRM version: %s\n", pNVRM_ID);
    len += sprintf(page+len, "GCC version:  %s\n", NV_COMPILER);
    
    return len;
}

int nv_kern_read_agpinfo(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    struct pci_dev *dev;
    char   *fw, *sba;
    u8     cap_ptr;
    u32    status, command, agp_rate;
    int    len = 0;
    
    nv_state_t *nv;
    nv = (nv_state_t *) data;

    if (nv) {
        dev = nv_find_pci_dev(nv);
        if (!dev)
            return 0;
    } else {
        dev = nvos_find_agp_by_class(PCI_CLASS_BRIDGE_HOST);
        if (!dev)
            return 0;

        len += sprintf(page+len, "Host Bridge: \t ");

#if defined(CONFIG_PCI_NAMES)
        len += sprintf(page+len, "%s\n", NV_PCI_DEVICE_NAME(dev));
#else
        len += sprintf(page+len, "PCI device %04x:%04x\n",
                dev->vendor, dev->device);
#endif
    }

    /* what can this AGP device do? */
    cap_ptr = nvos_find_agp_capability(dev);

    pci_read_config_dword(dev, cap_ptr + 4, &status);
    pci_read_config_dword(dev, cap_ptr + 8, &command);

    fw  = (status & 0x00000010) ? "Supported" : "Not Supported";
    sba = (status & 0x00000200) ? "Supported" : "Not Supported";

    len += sprintf(page+len, "Fast Writes: \t %s\n", fw);
    len += sprintf(page+len, "SBA: \t\t %s\n", sba);

    agp_rate = status & 0x7;
    if (status & 0x8) // agp 3.0
        agp_rate <<= 2;

    len += sprintf(page+len, "AGP Rates: \t %s%s%s%s\n",
            (agp_rate & 0x00000008) ? "8x " : "",
            (agp_rate & 0x00000004) ? "4x " : "",
            (agp_rate & 0x00000002) ? "2x " : "",
            (agp_rate & 0x00000001) ? "1x " : "");

    len += sprintf(page+len, "Registers: \t 0x%08x:0x%08x\n", status, command);

    return len;
}

int nv_kern_read_status(char *page, char **start, off_t off,
        int count, int *eof, void *data)
{
    struct pci_dev *dev;
    char   *fw, *sba, *drv;
    int    len = 0;
    u8     cap_ptr;
    u32    scratch;
    u32    status, command, agp_rate;

    nv_state_t *nv;
    nv = (nv_state_t *) data;

    dev = nvos_find_agp_by_class(PCI_CLASS_BRIDGE_HOST);
    if (!dev)
        return 0;
    cap_ptr = nvos_find_agp_capability(dev);

    pci_read_config_dword(dev, cap_ptr + 4, &status);
    pci_read_config_dword(dev, cap_ptr + 8, &command);

    dev = nvos_find_agp_by_class(PCI_CLASS_DISPLAY_VGA);
    if (!dev)
        return 0;
    cap_ptr = nvos_find_agp_capability(dev);

    pci_read_config_dword(dev, cap_ptr + 4, &scratch);
    status &= scratch;
    pci_read_config_dword(dev, cap_ptr + 8, &scratch);
    command &= scratch;

    if (NV_AGP_ENABLED(nv) && (command & 0x100)) {
        len += sprintf(page+len, "Status: \t Enabled\n");

        drv = NV_OSAGP_ENABLED(nv) ? "AGPGART" : "NVIDIA";
        len += sprintf(page+len, "Driver: \t %s\n", drv);

        // mask off agp rate. 
        // If this is agp 3.0, we need to shift the value
        agp_rate = command & 0x7;
        if (status & 0x8) // agp 3.0
            agp_rate <<= 2;

        len += sprintf(page+len, "AGP Rate: \t %dx\n", agp_rate);

        fw = (command & 0x00000010) ? "Enabled" : "Disabled";
        len += sprintf(page+len, "Fast Writes: \t %s\n", fw);

        sba = (command & 0x00000200) ? "Enabled" : "Disabled";
        len += sprintf(page+len, "SBA: \t\t %s\n", sba);
    } else {
        int agp_config = 0;

        len += sprintf(page+len, "Status: \t Disabled\n\n");

        /*
         * If we find AGP is disabled, but the RM registry indicates it
         * was requested, direct the user to the kernel log (we, or even
         * the kernel may have printed a warning/an error message).
         *
         * Note that the "XNvAGP" registry key reflects the user request
         * and overrides the RM "NvAGP" key, if present.
         */
        rm_read_registry_dword(nv, "NVreg", "NvAGP",  &agp_config);
        rm_read_registry_dword(nv, "NVreg", "XNvAGP", &agp_config);

        if (agp_config != NVOS_AGP_CONFIG_DISABLE_AGP && NV_AGP_FAILED(nv)) {
            len += sprintf(page+len,
                  "AGP initialization failed, please check the ouput  \n"
                  "of the 'dmesg' command and/or your system log file \n"
                  "for additional information on this problem.        \n");
        }
    }

    return len;
}


/***
 *** EXPORTS to rest of resman
 ***/

void* NV_API_CALL  nv_find_nv_mapping(
    nv_state_t    *nv,
    unsigned long  address
)
{
    nv_alloc_t *at;

    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), address, 
                        NV_ALLOC_TYPE_PCI);
    if (at && at->page_table)
    {
        // we've found the mapping and associated 'at' (in theory)
        // track down the actual page within this allocation and return
        // a kernel virtual mapping to it.
        int i;
        unsigned long offset;

        // save the page offset so we can add it to the returned address
        // page-align our address to make finding it a little easier
        offset = address & ~PAGE_MASK;
        address &= PAGE_MASK;

        // start off by tracking down which page within this allocation
        // we're looking at. do this by searching for the physical address
        // in our page table.
        for (i = 0; i < at->num_pages; i++)
        {
            if ((address == (unsigned long) at->page_table[i].phys_addr))
            {
                unsigned long retaddr = (unsigned long) at->page_table[i].phys_addr;

                // if we've allocated via vmalloc on a highmem system, the 
                // physical address may not be accessible via PAGE_OFFSET, 
                // that's ok, we have a simple linear pointer already.
                if (at->flags & NV_ALLOC_TYPE_VMALLOC)
                {
                    return (void *)((unsigned char *) at->key_mapping + (i << PAGE_SHIFT) + offset);
                }

                if (retaddr <= MAXMEM)
                {
                    return __va((retaddr + offset));
                }

                // ?? this may be a contiguous allocation, fall through
                // to below? or should I just check at->flag here?
            }
        }
    }

    // check if this was a contiguous kernel mapping...
    // make sure to change our physcal addr to a kernel virtual address
    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), 
                        (unsigned long) __va(address),
                        NV_ALLOC_TYPE_PCI | NV_ALLOC_TYPE_CONTIG);
    if (at)
        return at->key_mapping;

    return NULL;
}

ULONG NV_API_CALL nv_find_dma_mapping(
    nv_state_t    *nv,
    unsigned long  address
)
{
    nv_alloc_t *at = NULL;

    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), address, 
                        NV_ALLOC_TYPE_PCI);
    if (at && at->page_table)
    {
        // we've found the mapping and associated 'at' (in theory)
        // track down the actual page within this allocation and return
        // the agp aperture mapping to it (key_mapping should be the base
        // of this aperture mapping, so track down the page within that mapping)
        int i;
        unsigned long offset;

        // save the page offset so we can add it to the returned address
        // page-align our address to make finding it a little easier
        offset = address & ~PAGE_MASK;
        address &= PAGE_MASK;

        for (i = 0; i < at->num_pages; i++)
        {
            if (address == (unsigned long) at->page_table[i].phys_addr)
            {
                return (unsigned long)at->page_table[i].dma_addr + offset;
            }
        }
    }

    return 0;
}

/* For some newer AGP chipsets, such as the 460GX, the user's virtual address 
 * is not mapped directly to the agp aperture on the CPU's page tables. Instead,
 * they map to the underlying physical pages. This function is passed the
 * address of the underlying physical page (which is loaded into the GART) and
 * returns the agp aperture that the page is mapped to, so we can load that
 * page into the graphics card.
 * use the standard nvl_find_alloc to search on the physical page and rely on
 * the TYPE_AGP flag to differeniate it from a PCI allocation.
 * failure is fine, we may just be checking if a given page is agp
 */
void* NV_API_CALL nv_find_agp_kernel_mapping(
    nv_state_t    *nv,
    unsigned long  address
)
{
    nv_alloc_t *at = NULL;

    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), address, 
                        NV_ALLOC_TYPE_AGP);
    if (at && at->page_table)
    {
        // we've found the mapping and associated 'at' (in theory)
        // track down the actual page within this allocation and return
        // the agp aperture mapping to it (key_mapping should be the base
        // of this aperture mapping, so track down the page within that mapping)
        int i;
        for (i = 0; i < at->num_pages; i++)
        {
            if (address == (unsigned long) at->page_table[i].dma_addr)
            {
                return (void *)((unsigned long) at->key_mapping + 
                    (i * PAGE_SIZE));
            }
        }
    }

    return NULL;
}


#if defined(NVCPU_IA64)
#  define KERN_PAGE_MASK      _PFN_MASK
#else
#  define KERN_PAGE_MASK      PAGE_MASK
#endif

/* virtual address to physical page address */
static unsigned long
_get_phys_address(
    unsigned long address,
    int kern
)
{
    pgd_t *pg_dir;
    pmd_t *pg_mid_dir;
    pte_t *pte;
    unsigned long retval;

    if (kern) pg_dir = pgd_offset_k(address);
    else      pg_dir = pgd_offset(current->mm, address);

    if (!pg_dir || pgd_none(*pg_dir))
        goto failed;

    NV_PMD_OFFSET(address, pg_dir, pg_mid_dir);
    if (!NV_PMD_PRESENT(pg_mid_dir))
        goto failed;

    NV_PTE_OFFSET(address, pg_mid_dir, pte);

    if (!NV_PTE_PRESENT(pte))
        goto failed;

    retval = ((NV_PTE_VALUE(pte) & KERN_PAGE_MASK) | NV_MASK_OFFSET(address));

#if defined(NVCPU_X86_64) && defined(_PAGE_NX)
    // mask out the non-executable page bit for the true physical address
    retval &= ~_PAGE_NX;
#endif

    return retval;

  failed:
    return (unsigned long) NULL;
}

unsigned long NV_API_CALL nv_get_kern_phys_address(
    unsigned long address
)
{
    // make sure this address is a kernel pointer
    // IA64's memory layout is different from X86, at least in some cases.
    // instead of fine-tuning it, let's just bail, since we're really just 
    // trying to catch programming mistakes when debugging
#if defined(DEBUG) && !defined(CONFIG_X86_4G) && !defined(NVCPU_IA64)
    if (address < PAGE_OFFSET)
    {
        nv_printf(NV_DBG_WARNINGS,
            "NVRM: user address passed to get_kern_phys_address: 0x%lx\n",
            address);
        return 0;
    }
#endif

#if defined(NVCPU_IA64)
    if (address > __IA64_UNCACHED_OFFSET)
        return address - __IA64_UNCACHED_OFFSET;
#endif

    /* direct-mapped kernel address */
    if ((address > PAGE_OFFSET) && (address < VMALLOC_START))
        return __pa(address);

    return _get_phys_address(address, 1);
}

unsigned long NV_API_CALL nv_get_user_phys_address(
    unsigned long address
)
{
    // make sure this address is not a kernel pointer
    // IA64's memory layout is different from X86, at least in some cases.
    // instead of fine-tuning it, let's just bail, since we're really just 
    // trying to catch programming mistakes when debugging
#if defined(DEBUG) && !defined(CONFIG_X86_4G) && !defined(NVCPU_IA64)
    if (address >= PAGE_OFFSET)
    {
        nv_printf(NV_DBG_WARNINGS,
            "NVRM: kernel address passed to get_user_phys_address: 0x%lx\n",
            address);
        return 0;
    }
#endif

    return _get_phys_address(address, 0);
}


/* allocate memory for DMA push buffers */
int NV_API_CALL nv_alloc_pages(
    nv_state_t *nv,
    void **pAddress,
    unsigned int page_count,
    unsigned int agp_memory,
    unsigned int contiguous,
    unsigned int cached,
    unsigned int kernel,
    unsigned int class,
    void **priv_data
)
{
    nv_alloc_t *at;
    RM_STATUS rm_status = 0;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    nv_printf(NV_DBG_MEMINFO, "NVRM: nv_alloc_pages: %d pages\n", page_count);

    page_count = RM_PAGES_TO_OS_PAGES(page_count);
    at = nvos_create_alloc(nvl->dev, page_count);
    if (at == NULL)
        return RM_ERROR;

    at->class = class;
    at->flags |= NV_ALLOC_ENC_MAPPING(cached);

    if (agp_memory)
    {
        int offset;

        if (!NV_AGP_ENABLED(nv))
            goto failed;

        /* allocate agp-able memory */
        if (NV_OSAGP_ENABLED(nv))
        {
            /* agpgart will allocate all of the underlying memory */
            rm_status = KernAllocAGPPages(nv, pAddress, page_count, priv_data, &offset);
            if (rm_status)
                goto failed;

            at->priv_data = *priv_data;
            at->flags |= NV_ALLOC_TYPE_AGP;
            nvl_add_alloc(nvl, at);
        } else {
            /* use nvidia's nvagp support */
            if (nvos_malloc_pages(at))
                goto failed;

            at->class = class;

            // set our 'key' to the page_table. rm_alloc_agp_pages will call
            // nv_translate_agp_address below, which will look up pages using
            // the value of *pAddress as a key, then index into the page_table
            // once we're done with rm_alloc_agp_pages, we no longer need
            // this, and the 'key' will be replaced below
            *pAddress = at->page_table;
            at->key_mapping = at->page_table;
            at->flags |= NV_ALLOC_TYPE_AGP;

            /* the 'at' needs to be added before the alloc agp pages call */
            nvl_add_alloc(nvl, at);
            rm_status = rm_alloc_agp_pages(nv,
                                        pAddress,
                                        page_count,
                                        class,
                                        priv_data,
                                        &offset);
            if (rm_status)
            {
                nvl_remove_alloc(nvl, at);
                nvos_unlock_and_free_pages(NV_ATOMIC_READ(at->usage_count), at);
                goto failed;
            }
            at->priv_data = *priv_data;
        }
        // return the physical address of the allocation for mmap
        // in this case, 'physical address' is within the agp aperture
        *pAddress = (void *)(NV_UINTPTR_T)(nv->agp.address + (offset << PAGE_SHIFT));
    }
    else 
    {
        /* allocate general system memory */
        /* this is a little ugly */


        /* only use vmalloc if the kernel needs a linear pointer to noncontiguous, cached memory */
        if (kernel && !contiguous && NV_ALLOC_MAPPING_CACHED(at->flags))
        {
            int i;

#ifdef NV_SWIOTLB
            // unfortunately, this doesn't work too well yet.
            if (nv_swiotlb)
                goto failed;
#endif

            rm_status = RM_ERROR;
            NV_VMALLOC(*pAddress, page_count * PAGE_SIZE);
            if (*pAddress == NULL)
                goto failed;

            // load these pages into our page table for consistency across
            // allocation types (nv_kern_mmap will expect it).
            // make sure to get the physical address, since they may be remapped
            // to user space in nv_kern_mmap
            for (i = 0; i < at->num_pages; i++)
            {
                unsigned long addr = (unsigned long) *pAddress;
                at->page_table[i].virt_addr = addr + (i * PAGE_SIZE);
                at->page_table[i].phys_addr = nv_get_kern_phys_address((addr + (i * PAGE_SIZE)));
                at->page_table[i].dma_addr = at->page_table[i].phys_addr;

#if defined(NV_SG_MAP_BUFFERS)
                if (nv_sg_map_buffer(at->dev,
                                     &at->sg_list[i],
                                     __va(at->page_table[i].phys_addr),
                                     1))
                {
                    while (i >= 0)
                    {
                        nv_sg_unmap_buffer(at->dev, &at->sg_list[i], &at->page_table[i], 1);
                        i--;
                    }
                    NV_VFREE(*pAddress, at->num_pages * PAGE_SIZE);
                    goto failed;
                }
                at->page_table[i].dma_addr = at->sg_list[i].dma_address;
#endif
            }

            at->flags |= NV_ALLOC_TYPE_PCI | NV_ALLOC_TYPE_KERNEL | NV_ALLOC_TYPE_VMALLOC;
            nvl_add_alloc(nvl, at);
            rm_status = RM_OK;
        }
        /* if the kernel or user request contiguous memory, go for it
         * also if the kernel wants uncached memory, we need it to be contiguous and ioremap it
         */
        else if (contiguous || (kernel && !NV_ALLOC_MAPPING_CACHED(at->flags)))
        {
            int i;

            rm_status = os_alloc_contig_pages(pAddress, page_count * PAGE_SIZE);
            if (RM_OK != rm_status)
                goto failed;

            // use ioremap_nocache to get an uncached mapping to it
            if (kernel && !NV_ALLOC_MAPPING_CACHED(at->flags))
                /* XXX test on crush first */;

#if defined(NV_SG_MAP_BUFFERS)
            if (nv_sg_map_buffer(at->dev,
                                 &at->sg_list[0],
                                 *pAddress,
                                 page_count))
            {
                os_free_contig_pages(*pAddress, page_count * PAGE_SIZE);
                goto failed;
            }
#endif

            // load these pages into our page table for consistency across
            // allocation types (nv_kern_mmap will expect it).
            // make sure to get the physical address, since they may be remapped
            // to user space in nv_kern_mmap
            for (i = 0; i < at->num_pages; i++)
            {
                unsigned long addr = (unsigned long) *pAddress;
                at->page_table[i].virt_addr = addr + (i * PAGE_SIZE);
                at->page_table[i].phys_addr = __pa((addr + (i * PAGE_SIZE)));
                at->page_table[i].dma_addr = at->page_table[i].phys_addr;

#if defined(NV_SG_MAP_BUFFERS)
                nv_sg_load(&at->sg_list[i], &at->page_table[i], 1);
#endif
            }
            *pAddress = (void *) at->page_table[0].virt_addr;

            at->flags |= NV_ALLOC_TYPE_PCI | NV_ALLOC_TYPE_CONTIG;
            if (kernel)
                at->flags |= NV_ALLOC_TYPE_KERNEL;
            nvl_add_alloc(nvl, at);
        }
        /* this path is only for user-requested allocations */
        else
        if (!kernel)
        {
            if (nvos_malloc_pages(at))
                goto failed;

            /* must be page-aligned or mmap will fail
             * so use the first page, which is page-aligned. this way, our 
             * allocated page table does not need to be page-aligned
             */
            *pAddress = (void *) at->page_table[0].phys_addr;
            at->flags |= NV_ALLOC_TYPE_PCI;
            nvl_add_alloc(nvl, at);
        } else
            goto failed;
    }

    at->key_mapping = *pAddress;
    NV_ATOMIC_INC(at->usage_count);

    return RM_OK;

failed:
    nvos_free_alloc(at);

    return -1;
}

#define NV_FAILED_TO_FIND_AT(nv, addr) \
    { \
        nv_up(nvl->at_lock); \
        nv_printf(NV_DBG_ERRORS, "NVRM: couldn't find alloc for 0x%p\n", addr); \
        return -1; \
    }

int NV_API_CALL nv_free_pages(
    nv_state_t *nv,
    void **pAddress,
    unsigned int page_count,
    unsigned int agp_memory,
    void *priv_data
)
{
    int rmStatus = 0;
    nv_alloc_t *at;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    page_count = RM_PAGES_TO_OS_PAGES(page_count);
    nv_printf(NV_DBG_MEMINFO, "NVRM: nv_free_pages: 0x%x 0x%x\n",
        *pAddress, page_count);

    if (agp_memory)
    {
        if (!NV_AGP_ENABLED(nv))
            return -1;

        /* only lock ldata while removing 'at' from the list */
        nv_down(nvl->at_lock);
        at = nvl_find_alloc(nvl, (unsigned long) *pAddress, NV_ALLOC_TYPE_AGP);
        if (at == NULL)
            NV_FAILED_TO_FIND_AT(nv, *pAddress);
        if (at->num_pages != page_count)
            NV_FAILED_TO_FIND_AT(nv, *pAddress);
        nvl_remove_alloc(nvl, at);
        nv_up(nvl->at_lock);

        NV_ATOMIC_DEC(at->usage_count);

        if (NV_OSAGP_ENABLED(nv))
        {
            rmStatus = KernFreeAGPPages(nv, pAddress, priv_data);
        } else {
            rmStatus = rm_free_agp_pages(nv, pAddress, priv_data);
            if (rmStatus == RM_OK)
            {
                nvos_unlock_and_free_pages(NV_ATOMIC_READ(at->usage_count), at);
            }
        }
    } else {
        /* only lock ldata while removing 'at' from the list */
        nv_down(nvl->at_lock);
        at = nvl_find_alloc(nvl, (unsigned long) *pAddress, NV_ALLOC_TYPE_PCI);
        if (at == NULL)
            NV_FAILED_TO_FIND_AT(nv, *pAddress);
        if (at->num_pages != page_count)
            NV_FAILED_TO_FIND_AT(nv, *pAddress);
        nvl_remove_alloc(nvl, at);
        nv_up(nvl->at_lock);

        NV_ATOMIC_DEC(at->usage_count);

        if (at->flags & NV_ALLOC_TYPE_CONTIG)
        {
#if defined(NV_SG_MAP_BUFFERS)
            nv_sg_unmap_buffer(at->dev, at->sg_list, at->page_table, at->num_pages);
            at->key_mapping = (void *)at->page_table[0].virt_addr;
#endif
            os_free_contig_pages(at->key_mapping, at->num_pages * PAGE_SIZE);
        }
        else if (at->flags & NV_ALLOC_TYPE_VMALLOC)
        {
#if defined(NV_SG_MAP_BUFFERS)
            int i;
            for (i = 0; i < at->num_pages; i++)
            {
                nv_sg_unmap_buffer(at->dev, &at->sg_list[i], &at->page_table[i], 1);
            }
#endif
            NV_VFREE(at->key_mapping, at->num_pages * PAGE_SIZE);
        }
        else
        {
            nvos_unlock_and_free_pages(NV_ATOMIC_READ(at->usage_count), at);
        }
    }

    if (NV_ATOMIC_READ(at->usage_count) == 0)
        nvos_free_alloc(at);

    return rmStatus;
}


/* avoid compiler warnings on UP kernels, 
 * when spinlock macros are defined away 
 */
#define NO_COMPILER_WARNINGS(nvl) \
    if (nvl == NULL) return


static void nv_lock_init_locks
( 
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    NO_COMPILER_WARNINGS(nvl);

    spin_lock_init(&nvl->rm_lock);

    sema_init(&nvl->ldata_lock, 1);
    sema_init(&nvl->at_lock, 1);
    NV_ATOMIC_SET(nvl->usage_count, 0);

    nvl->rm_lock_cpu = -1;
    nvl->rm_lock_count = 0;
}

void NV_API_CALL nv_lock_rm(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);
    NO_COMPILER_WARNINGS(nvl);

    if (nvl->rm_lock_cpu == smp_processor_id())
    {
        nvl->rm_lock_count++;
        return;
    }

    spin_unlock_wait(&nvl->rm_lock);
    spin_lock_irq(&nvl->rm_lock);

   nvl->rm_lock_cpu = smp_processor_id();
   nvl->rm_lock_count = 1;
}

void NV_API_CALL nv_unlock_rm(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);
    NO_COMPILER_WARNINGS(nvl);

    if (--nvl->rm_lock_count)
        return;

    nvl->rm_lock_cpu = -1;
    spin_unlock_irq(&nvl->rm_lock);
}

/*
** post the event
*/
 
void NV_API_CALL nv_post_event(
    nv_state_t *nv,
    nv_event_t *event,
    U032        handle,
    U032        index
)
{
    struct file *file = (struct file *) event->file;
    nv_file_private_t *nvfp = NV_GET_NVFP(file);
    unsigned long eflags;

    nv_printf(NV_DBG_EVENTINFO, "NVRM: posting event on 0x%x:0x%x\n",
        event, nvfp);

    nv_lock_irq(nvfp->fp_lock, eflags);

    if (nvfp->num_events == NV_EVENT_FIFO_SIZE)
    {
        wake_up_interruptible(&nvfp->waitqueue);
        nv_unlock_irq(nvfp->fp_lock, eflags);
        return;
    }

    // copy the event into the queue
    nvfp->event_fifo[nvfp->put] = *event;

    // set the handle for this event
    nvfp->event_fifo[nvfp->put].hObject = handle;
    nvfp->event_fifo[nvfp->put].index   = index;
    
    nvfp->num_events++;
    nvfp->put++;
    if (nvfp->put >= NV_EVENT_FIFO_SIZE)
        nvfp->put = 0;

    wake_up_interruptible(&nvfp->waitqueue);
    nv_unlock_irq(nvfp->fp_lock, eflags);
}

int NV_API_CALL nv_get_event(
    nv_state_t *nv,
    VOID *void_file,
    nv_event_t *event,
    U032 *more_events
)
{
    struct file *file = (struct file *) void_file;
    nv_file_private_t *nvfp = NV_GET_NVFP(file);
    unsigned long eflags;

    nv_lock_irq(nvfp->fp_lock, eflags);
    if (nvfp->num_events == 0)
    {
        nv_unlock_irq(nvfp->fp_lock, eflags);
        return -1;
    }

    *event = nvfp->event_fifo[nvfp->get];
    nvfp->num_events--;
    nvfp->get++;
    if (nvfp->get >= NV_EVENT_FIFO_SIZE)
        nvfp->get = 0;

    if (more_events)
        *more_events = nvfp->num_events;

    nv_printf(NV_DBG_EVENTINFO, "NVRM: returning event: 0x%x\n", event);
    nv_printf(NV_DBG_EVENTINFO, "NVRM:     hParent: 0x%x\n", event->hParent);
    nv_printf(NV_DBG_EVENTINFO, "NVRM:     hObject: 0x%x\n", event->hObject);
    nv_printf(NV_DBG_EVENTINFO, "NVRM:     file:    0x%p\n", event->file);
    nv_printf(NV_DBG_EVENTINFO, "NVRM:     fd:      %d\n", event->fd);
    if (more_events)
        nv_printf(NV_DBG_EVENTINFO, "NVRM: more events: %d\n", *more_events);

    nv_unlock_irq(nvfp->fp_lock, eflags);

    return 0;
}


int NV_API_CALL nv_agp_init(
    nv_state_t *nv,
    VOID **phys_start,
    VOID **linear_start,
    VOID *agp_limit,
    U032 config         /* passed in from XF86Config file */
)
{
    U032 status = 1;
    static int old_error = 0;

    if (NV_AGP_ENABLED(nv))
        return -1;

    if (config == NVOS_AGP_CONFIG_DISABLE_AGP)
    {
        nv->agp_config = NVOS_AGP_CONFIG_DISABLE_AGP;
        nv->agp_status = NV_AGP_STATUS_DISABLED;
        return 0;
    }

    nv_printf(NV_DBG_SETUP, "NVRM: nv_agp_init\n");

    nv->agp_config = NVOS_AGP_CONFIG_DISABLE_AGP;
    nv->agp_status = NV_AGP_STATUS_FAILED;

    if (config & NVOS_AGP_CONFIG_OSAGP)
    {
        status = KernInitAGP(nv, phys_start, linear_start, agp_limit);

        /* if enabling agpgart was successfull, register it,
         * and check about overrides
         */
        if (status == 0)
        {
            nv->agp_config = NVOS_AGP_CONFIG_OSAGP;
            nv->agp_status = NV_AGP_STATUS_ENABLED;

            /* make sure we apply our overrides in this case */
            rm_update_agp_config(nv);
        }

        if (status == 1 && !(config & NVOS_AGP_CONFIG_NVAGP) && !old_error)
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: unable to initialize the Linux AGPGART driver, please \n"
                "NVRM: verify you configured your kernel to include support  \n"
                "NVRM: for AGPGART (either statically linked, or as a kernel \n"
                "NVRM: module). Please also make sure you selected support   \n"
                "NVRM: for your AGP chipset.                                 \n");
#if defined(KERNEL_2_6)
            nv_printf(NV_DBG_ERRORS,
                "NVRM:                                                       \n"
                "NVRM: note that as of Linux 2.6 AGPGART, all chipset/vendor \n"
                "NVRM: drivers are split into independent modules; make sure \n"
                "NVRM: the correct one is loaded for your chipset.           \n");
#endif
            old_error = 1;
        }

        /* if agpgart is loaded, but we failed to initialize it,
         * we'd better not attempt nvagp, or we're likely to lock
         * the machine.
         */
        if (status < 0)
            return status;
    }

    /* we're either explicitly not using agpgart,
     * or trying to use agpgart failed
     * make sure the user did not specify "use agpgart only"
     */
    if ( (!NV_AGP_ENABLED(nv)) && (config & NVOS_AGP_CONFIG_NVAGP) )
    {
        /* make sure the user does not have agpgart loaded */
        if (inter_module_get("drm_agp")) {
            inter_module_put("drm_agp");
            nv_printf(NV_DBG_WARNINGS, "NVRM: not using NVAGP, AGPGART is loaded!!\n");
        } else {
#if defined(CONFIG_X86_64) && defined(CONFIG_GART_IOMMU)
            nv_printf(NV_DBG_WARNINGS,
                "NVRM: not using NVAGP, kernel was compiled with GART_IOMMU support!!\n");
#else
            status = rm_init_agp(nv);
            if (status == RM_OK)
            {
                nv->agp_config = NVOS_AGP_CONFIG_NVAGP;
                nv->agp_status = NV_AGP_STATUS_ENABLED;
            }
#endif
        }
    }

    if (NV_AGP_ENABLED(nv))
        old_error = 0; /* report new errors */

    nv_printf(NV_DBG_SETUP, 
        "NVRM: agp_init finished with status 0x%x and config %d\n",
        status, nv->agp_config);

    return status;
}

int NV_API_CALL nv_agp_teardown(
    nv_state_t *nv
)
{
    U032 status = 1;

    nv_printf(NV_DBG_SETUP, "NVRM: nv_agp_teardown\n");

    /* little sanity check won't hurt */
    if (!NV_AGP_ENABLED(nv))
        return -1;

    if (NV_OSAGP_ENABLED(nv))
        status = KernTeardownAGP(nv);
    else if (NV_NVAGP_ENABLED(nv))
        status = rm_teardown_agp(nv);

    nv->agp_config = NVOS_AGP_CONFIG_DISABLE_AGP;
    nv->agp_status = NV_AGP_STATUS_DISABLED;

    nv_printf(NV_DBG_SETUP, "NVRM: teardown finished with status 0x%x\n", 
        status);

    return status;
}

/*
 * this takes a virtual address related to agp and returns the physical address
 * this part is a bit hacky.. nvagp expects a kernel-linear address mapping,
 * but we don't have that. So, what's really passed to us is a "fake linear"
 * address, based at our page_table. nvl_find_agp_alloc() knows this, and will
 * return the 'at' that has the page_table. From there, we deconstruct this
 * 'fake address' to get a "page index" into the virtual address. We use this
 * index to grab a page from our page_table and return that physical address.
 * (if the linear mapping was real, this is the page that would have been
 * addressed by the virtual address. I know, confusing).
 */
int NV_API_CALL nv_agp_translate_address(
    nv_state_t *nv,
    void       *base,
    U032        index,
    U032       *paddr
)
{
    nv_alloc_t *at;

    /* find the 'at' */
    at = nvl_find_alloc(NV_GET_NVL_FROM_NV_STATE(nv), 
                        (unsigned long) base, NV_ALLOC_TYPE_AGP);
    if (at == NULL)
        return -1;

    if (index > at->num_pages)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: translate_address: ",
            "at has inconsistent number of pages\n");
        return -1;
    }

    /* get the physical address of this page */
    *paddr = (U032) ((NV_UINTPTR_T)at->page_table[index].dma_addr);

    return 0x0;
}


int NV_API_CALL nv_int10h_call(
    nv_state_t *nv,
    U032 *eax,
    U032 *ebx,
    U032 *ecx,
    U032 *edx,
    VOID *buffer
)
{
    return -1;
}

/* set a timer to go off every second */
int NV_API_CALL nv_start_rc_timer(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    if (nv->rc_timer_enabled)
        return -1;

    nv_printf(NV_DBG_INFO, "NVRM: initializing rc timer\n");
    init_timer(&nvl->rc_timer);
    nvl->rc_timer.function = nv_kern_rc_timer;
    nvl->rc_timer.data = (unsigned long) nv;
    nv->rc_timer_enabled = 1;
    mod_timer(&nvl->rc_timer, jiffies + HZ); /* set our timeout for 1 second */
    nv_printf(NV_DBG_INFO, "NVRM: rc timer initialized\n");

    return 0;
}

int NV_API_CALL nv_stop_rc_timer(
    nv_state_t *nv
)
{
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);

    if (!nv->rc_timer_enabled)
        return -1;

    nv_printf(NV_DBG_INFO, "NVRM: stopping rc timer\n");
    nv->rc_timer_enabled = 0;
    del_timer(&nvl->rc_timer);
    nv_printf(NV_DBG_INFO, "NVRM: rc timer stopped\n");

    return 0;
}

/* make sure the pci_driver called probe for all of our devices.
 * we've seen cases where rivafb claims the device first and our driver
 * doesn't get called.
 */
static int
nvos_count_devices(void)
{
    struct pci_dev *dev = NULL;
    int count = 0;

    dev = pci_find_class(PCI_CLASS_DISPLAY_VGA << 8, dev);
    while (dev)
    {
        if ((dev->vendor == 0x10de) && (dev->device >= 0x20))
            count++;
        dev = pci_find_class(PCI_CLASS_DISPLAY_VGA << 8, dev);
    }
    return count;
}

/* find nvidia devices and set initial state */
int
nv_kern_probe
(
    struct pci_dev *dev,
    const struct pci_device_id *id_table
)
{
    nv_state_t *nv;
    nv_linux_state_t *nvl;

    if ((dev->vendor != 0x10de) || (dev->device < 0x20) || 
        (dev->class != (PCI_CLASS_DISPLAY_VGA << 8)))
    {
        return -1;
    }

    if (num_nv_devices == NV_MAX_DEVICES)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: maximum device number (%d) reached!\n", num_nv_devices);
        return -1;
    }

    /* initialize bus-dependent config state */
    nvl = &nv_linux_devices[num_nv_devices];
    nv  = NV_STATE_PTR(nvl);

    nvl->dev          = dev;
    nv->vendor_id     = dev->vendor;
    nv->device_id     = dev->device;
    nv->os_state      = (void *) nvl;
    nv->bus           = dev->bus->number;
    nv->slot          = PCI_SLOT(dev->devfn);

    nv_lock_init_locks(nv);
    

    nv->bar.regs.address = dev->resource[0].start;
    nv->bar.regs.size    = dev->resource[0].end - dev->resource[0].start + 1;

    nv->bar.fb.address   = dev->resource[1].start;
    nv->bar.fb.size      = dev->resource[1].end - dev->resource[1].start + 1;
    
    nv->interrupt_line = dev->irq;

    /* check common error condition */
    if (nv->interrupt_line == 0)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: Can't find an IRQ for your NVIDIA card!  \n");
        nv_printf(NV_DBG_ERRORS, "NVRM: Please check your BIOS settings.         \n");
        nv_printf(NV_DBG_ERRORS, "NVRM: [Plug & Play OS   ] should be set to NO  \n");
        nv_printf(NV_DBG_ERRORS, "NVRM: [Assign IRQ to VGA] should be set to YES \n");
        return -1;
    }

    /* sanity check the IO apertures */
    if ( (nv->bar.regs.address == 0) || (nv->bar.regs.size == 0) ||
         (nv->bar.fb.address   == 0) || (nv->bar.fb.size   == 0))
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: The IO regions for your NVIDIA card are invalid.\n");
        nv_printf(NV_DBG_ERRORS, "NVRM: Your system BIOS may have misconfigured your graphics card.\n");

        if ((nv->bar.regs.address == 0) || (nv->bar.regs.size == 0))
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: bar0 (registers) appears to be wrong: 0x%x 0x%x\n",
                nv->bar.regs.address, nv->bar.regs.size);
        }

        if ((nv->bar.fb.address == 0) || (nv->bar.fb.size == 0))
        {
            nv_printf(NV_DBG_ERRORS,
                "NVRM: bar1 (framebuffer) appears to be wrong: 0x%x 0x%x\n",
                nv->bar.fb.address, nv->bar.fb.size);
        }

        /* Clear out the data */
        os_mem_set(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }

    // request ownership of our bars
    // keeps other drivers from banging our registers.
    // only do this for registers, as vesafb requests our framebuffer and will
    // keep us from working properly
    if (!request_mem_region(nv->bar.regs.address, nv->bar.regs.size, "nvidia"))
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: pci_request_regions failed, aborting\n");

        /* Clear out the data */
        os_mem_set(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }

    // enable io, mem, and bus-mastering in pci config space
    if (pci_enable_device(dev) != 0)
    {
        nv_printf(NV_DBG_ERRORS,
            "NVRM: pci_enable_device failed, aborting\n");

        pci_release_regions(dev);
        os_mem_set(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }
    pci_set_master(nvl->dev);

#if defined(NV_BUILD_NV_PAT_SUPPORT)
    if (nvos_find_pci_express_capability(nvl->dev))
    {
        __nv_enable_pat_support();
    }
#endif

#if defined(NV_MAP_REGISTERS_EARLY)
    nv->bar.regs.map = os_map_kernel_space(nv->bar.regs.address,
                                           nv->bar.regs.size,
                                           NV_MEMORY_UNCACHED);
    nv->bar.regs.map_u = (nv_phwreg_t) nv->bar.regs.map;
    if (nv->bar.regs.map == NULL)
    {
        nv_printf(NV_DBG_ERRORS, "NVRM: failed to map registers!!\n");

        pci_release_regions(dev);

        /* Clear out the data */
        os_mem_set(nvl, 0, sizeof(nv_linux_state_t));

        return -1;
    }
    nv->flags |= NV_FLAG_MAP_REGS_EARLY;
#endif

    nv_printf(NV_DBG_INFO, "NVRM: %02x:%02x.%x %04x:%04x - 0x%08x [size=%dM]\n",
            nv->bus, nv->slot, PCI_FUNC(dev->devfn),
            nv->vendor_id, nv->device_id, nv->bar.regs.address,
            nv->bar.regs.size / (1024 * 1024));
    nv_printf(NV_DBG_INFO, "NVRM: %02x:%02x.%x %04x:%04x - 0x%08x [size=%dM]\n",
            nv->bus, nv->slot, PCI_FUNC(dev->devfn),
            nv->vendor_id, nv->device_id, nv->bar.fb.address,
            nv->bar.fb.size / (1024 * 1024));

    num_nv_devices++;

    return 0;
}

int NV_API_CALL nv_no_incoherent_mappings
(
    void
)
{
#ifdef NV_CHANGE_PAGE_ATTR_PRESENT
    return 1;
#else
    return 0;
#endif
}

#if defined(NV_PM_SUPPORT_ACPI)

int
nv_acpi_event
(
    struct pci_dev *dev, 
    u32 state
)
{
    nv_state_t *nv;
    nv_linux_state_t *lnv = NULL;
    int status = RM_OK, i;

    nv_printf(NV_DBG_INFO, "NVRM: nv_acpi_event: %d\n", state);

    for (i = 0; i < NV_MAX_DEVICES; i++)
    {
        if (nv_linux_devices[i].dev == dev)
        {
            lnv = &nv_linux_devices[i];
            break;
        }
    }

    if ((!lnv) || (lnv->dev != dev))
    {
        nv_printf(NV_DBG_WARNINGS, "NVRM: ACPI: invalid device!\n");
        return -1;
    }

    nv = NV_STATE_PTR(lnv);
    if (nv->pdev == NULL)
    {
        nv_printf(NV_DBG_WARNINGS, "NVRM: ACPI: device not initialized!\n");
        return -1;
    }

    switch (state)
    {
        case PM_SAVE_STATE:
            nv_printf(NV_DBG_INFO, "NVRM: ACPI: received standby event\n");
            status = rm_power_management(nv, 0, NV_PM_ACPI_STANDBY);
            break;

        case PM_RESUME:
            nv_printf(NV_DBG_INFO, "NVRM: ACPI: received resume event\n");
            status = rm_power_management(nv, 0, NV_PM_ACPI_RESUME);
            break;

        default:
            nv_printf(NV_DBG_WARNINGS, "NVRM: ACPI: unsupported event: %d\n", state);
            return -1;
    }

    if (status != RM_OK)
        nv_printf(NV_DBG_ERRORS, "NVRM: ACPI: failed event: %d\n", state);

    return status;
}

int
nv_kern_acpi_standby
(
    struct pci_dev *dev, 
    u32 state
)
{
    return nv_acpi_event(dev, state);
}

int
nv_kern_acpi_resume
(
    struct pci_dev *dev
)
{
    return nv_acpi_event(dev, PM_RESUME);
}

#endif
